# App.transport creates Kafka consumer and producer.
import abc
import asyncio
import ssl
import typing
from typing import (
    AbstractSet,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    ClassVar,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    no_type_check,
)

from mode import Seconds, ServiceT
from yarl import URL

from .core import HeadersArg
from .topics import TopicT
from .tuples import FutureMessage, Message, RecordMetadata, TP

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
else:
    class _AppT: ...  # noqa

__all__ = [
    'ConsumerCallback',
    'TPorTopicSet',
    'PartitionsRevokedCallback',
    'PartitionsAssignedCallback',
    'PartitionerT',
    'ConsumerT',
    'ProducerT',
    'ConductorT',
    'TransactionManagerT',
    'TransportT',
]

#: Callback called by :class:`faust.transport.base.Consumer` whenever
#: a message is received.
ConsumerCallback = Callable[[Message], Awaitable]

#: Argument to Consumer.commit to specify topics/tps to commit.
TPorTopic = Union[str, TP]
TPorTopicSet = AbstractSet[TPorTopic]

#: Callback (:keyword:`async def`) called when consumer partitions are revoked.
PartitionsRevokedCallback = Callable[[Set[TP]], Awaitable[None]]

#: Callback (:keyword:`async def`) called when consumer
#: partitions are assigned.
PartitionsAssignedCallback = Callable[[Set[TP]], Awaitable[None]]

PartitionerT = Callable[
    [Optional[bytes],   # key to hash (or None)
     Sequence[int],     # all partitions
     Sequence[int]],    # available partitions
    int,
]


class ProducerBufferT(ServiceT):

    max_messages: int

    pending: asyncio.Queue

    @abc.abstractmethod
    def put(self, fut: FutureMessage) -> None:
        ...

    @abc.abstractmethod
    async def flush(self) -> None:
        ...

    @abc.abstractmethod
    async def flush_atmost(self, n: int) -> int:
        ...

    @abc.abstractmethod
    async def wait_until_ebb(self) -> None:
        ...

    @property
    @abc.abstractmethod
    def size(self) -> int:
        ...


class ProducerT(ServiceT):

    #: The transport that created this Producer.
    transport: 'TransportT'

    buffer: ProducerBufferT

    client_id: str
    linger_ms: int
    max_batch_size: int
    acks: int
    max_request_size: int
    compression_type: Optional[str]
    ssl_context: Optional[ssl.SSLContext]
    partitioner: Optional[PartitionerT]
    request_timeout: float

    @abc.abstractmethod
    def __init__(self, transport: 'TransportT',
                 loop: asyncio.AbstractEventLoop = None,
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    async def send(self, topic: str, key: Optional[bytes],
                   value: Optional[bytes],
                   partition: Optional[int],
                   timestamp: Optional[float],
                   headers: Optional[HeadersArg],
                   *,
                   transactional_id: str = None) -> Awaitable[RecordMetadata]:
        ...

    @abc.abstractmethod
    def send_soon(self, fut: FutureMessage) -> None:
        ...

    @abc.abstractmethod
    async def send_and_wait(self, topic: str, key: Optional[bytes],
                            value: Optional[bytes],
                            partition: Optional[int],
                            timestamp: Optional[float],
                            headers: Optional[HeadersArg],
                            *,
                            transactional_id: str = None) -> RecordMetadata:
        ...

    @abc.abstractmethod
    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 1000.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        ...

    @abc.abstractmethod
    def key_partition(self, topic: str, key: bytes) -> TP:
        ...

    @abc.abstractmethod
    async def flush(self) -> None:
        ...

    @abc.abstractmethod
    async def begin_transaction(self, transactional_id: str) -> None:
        ...

    @abc.abstractmethod
    async def commit_transaction(self, transactional_id: str) -> None:
        ...

    @abc.abstractmethod
    async def abort_transaction(self, transactional_id: str) -> None:
        ...

    @abc.abstractmethod
    async def stop_transaction(self, transactional_id: str) -> None:
        ...

    @abc.abstractmethod
    async def maybe_begin_transaction(self, transactional_id: str) -> None:
        ...

    @abc.abstractmethod
    async def commit_transactions(
            self,
            tid_to_offset_map: Mapping[str, Mapping[TP, int]],
            group_id: str,
            start_new_transaction: bool = True) -> None:
        ...

    @abc.abstractmethod
    def supports_headers(self) -> bool:
        ...


class TransactionManagerT(ProducerT):
    consumer: 'ConsumerT'
    producer: 'ProducerT'

    @abc.abstractmethod
    def __init__(self,
                 transport: 'TransportT',
                 loop: asyncio.AbstractEventLoop = None,
                 *,
                 consumer: 'ConsumerT',
                 producer: 'ProducerT',
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    async def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def on_rebalance(self,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def commit(self, offsets: Mapping[TP, int],
                     start_new_transaction: bool = True) -> bool:
        ...

    async def begin_transaction(self, transactional_id: str) -> None:
        raise NotImplementedError()

    async def commit_transaction(self, transactional_id: str) -> None:
        raise NotImplementedError()

    async def abort_transaction(self, transactional_id: str) -> None:
        raise NotImplementedError()

    async def stop_transaction(self, transactional_id: str) -> None:
        raise NotImplementedError()

    async def maybe_begin_transaction(self, transactional_id: str) -> None:
        raise NotImplementedError()

    async def commit_transactions(
            self,
            tid_to_offset_map: Mapping[str, Mapping[TP, int]],
            group_id: str,
            start_new_transaction: bool = True) -> None:
        raise NotImplementedError()


class SchedulingStrategyT:

    @abc.abstractmethod
    def __init__(self) -> None:
        ...

    @abc.abstractmethod
    def iterate(self, records: Mapping[TP, List]) -> Iterator[Tuple[TP, Any]]:
        ...


class ConsumerT(ServiceT):

    #: The transport that created this Consumer.
    transport: 'TransportT'

    transactions: TransactionManagerT

    #: How often we commit topic offsets.
    #: See :setting:`broker_commit_interval`.
    commit_interval: float

    #: Set of topic names that are considered "randomly assigned".
    #: This means we don't crash if it's not part of our assignment.
    #: Used by e.g. the leader assignor service.
    randomly_assigned_topics: Set[str]

    in_transaction: bool

    scheduler: SchedulingStrategyT

    @abc.abstractmethod
    def __init__(self,
                 transport: 'TransportT',
                 callback: ConsumerCallback,
                 on_partitions_revoked: PartitionsRevokedCallback,
                 on_partitions_assigned: PartitionsAssignedCallback,
                 *,
                 commit_interval: float = None,
                 loop: asyncio.AbstractEventLoop = None,
                 **kwargs: Any) -> None:
        self._on_partitions_revoked: PartitionsRevokedCallback
        self._on_partitions_assigned: PartitionsAssignedCallback

    @abc.abstractmethod
    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 1000.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        ...

    @abc.abstractmethod
    async def subscribe(self, topics: Iterable[str]) -> None:
        ...

    @abc.abstractmethod
    @no_type_check
    async def getmany(self,
                      timeout: float) -> AsyncIterator[Tuple[TP, Message]]:
        ...

    @abc.abstractmethod
    def track_message(self, message: Message) -> None:
        ...

    @abc.abstractmethod
    async def perform_seek(self) -> None:
        ...

    @abc.abstractmethod
    def ack(self, message: Message) -> bool:
        ...

    @abc.abstractmethod
    async def wait_empty(self) -> None:
        ...

    @abc.abstractmethod
    def assignment(self) -> Set[TP]:
        ...

    @abc.abstractmethod
    def highwater(self, tp: TP) -> int:
        ...

    @abc.abstractmethod
    async def stop_flow(self) -> None:
        ...

    @abc.abstractmethod
    def resume_flow(self) -> None:
        ...

    @abc.abstractmethod
    def pause_partitions(self, tps: Iterable[TP]) -> None:
        ...

    @abc.abstractmethod
    def resume_partitions(self, tps: Iterable[TP]) -> None:
        ...

    @abc.abstractmethod
    async def position(self, tp: TP) -> Optional[int]:
        ...

    @abc.abstractmethod
    async def seek(self, partition: TP, offset: int) -> None:
        ...

    @abc.abstractmethod
    async def seek_wait(self, partitions: Mapping[TP, int]) -> None:
        ...

    @abc.abstractmethod
    async def commit(self,
                     topics: TPorTopicSet = None,
                     start_new_transaction: bool = True) -> bool:
        ...

    @abc.abstractmethod
    async def on_task_error(self, exc: BaseException) -> None:
        ...

    @abc.abstractmethod
    async def earliest_offsets(self, *partitions: TP) -> Mapping[TP, int]:
        ...

    @abc.abstractmethod
    async def highwaters(self, *partitions: TP) -> Mapping[TP, int]:
        ...

    @abc.abstractmethod
    def topic_partitions(self, topic: str) -> Optional[int]:
        ...

    @abc.abstractmethod
    def key_partition(self,
                      topic: str,
                      key: Optional[bytes],
                      partition: int = None) -> Optional[int]:
        ...

    @abc.abstractmethod
    def close(self) -> None:
        ...

    @abc.abstractmethod
    def verify_recovery_event_path(self, now: float, tp: TP) -> None:
        ...

    @property
    @abc.abstractmethod
    def unacked(self) -> Set[Message]:
        ...

    @abc.abstractmethod
    def request_rejoin(self) -> None:
        ...


class ConductorT(ServiceT, MutableSet[TopicT]):

    # The topic conductor delegates messages from the Consumer
    # to the various Topic instances subscribed to a topic.

    app: _AppT

    @abc.abstractmethod
    def __init__(self, app: _AppT, **kwargs: Any) -> None:
        self.on_message: ConsumerCallback

    @abc.abstractmethod
    async def on_client_only_start(self) -> None:
        ...

    @abc.abstractmethod
    def acks_enabled_for(self, topic: str) -> bool:
        ...

    @abc.abstractmethod
    async def commit(self, topics: TPorTopicSet) -> bool:
        ...

    @abc.abstractmethod
    async def wait_for_subscriptions(self) -> None:
        ...

    @abc.abstractmethod
    async def maybe_wait_for_subscriptions(self) -> None:
        ...

    @abc.abstractmethod
    async def on_partitions_assigned(self, assigned: Set[TP]) -> None:
        ...

    @property
    @abc.abstractmethod
    def acking_topics(self) -> Set[str]:
        ...


class TransportT(abc.ABC):

    #: The Consumer class used for this type of transport.
    Consumer: ClassVar[Type[ConsumerT]]

    #: The Producer class used for this type of transport.
    Producer: ClassVar[Type[ProducerT]]

    #: The TransactionManager class used for managing multiple transactions.
    TransactionManager: ClassVar[Type[TransactionManagerT]]

    #: The Conductor class used to delegate messages from Consumer to streams.
    Conductor: ClassVar[Type[ConductorT]]

    #: The Fetcher service used for this type of transport.
    Fetcher: ClassVar[Type[ServiceT]]

    #: The :class:`faust.App` that created this transport.
    app: _AppT

    #: The URL to use for this transport (e.g. kafka://localhost).
    url: List[URL]

    #: String identifying the underlying driver used for this transport.
    #: E.g. for :pypi:`aiokafka` this could be ``aiokafka 0.4.1``.
    driver_version: str

    loop: asyncio.AbstractEventLoop

    @abc.abstractmethod
    def __init__(self,
                 url: List[URL],
                 app: _AppT,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        ...

    @abc.abstractmethod
    def create_consumer(self, callback: ConsumerCallback,
                        **kwargs: Any) -> ConsumerT:
        ...

    @abc.abstractmethod
    def create_producer(self, **kwargs: Any) -> ProducerT:
        ...

    @abc.abstractmethod
    def create_transaction_manager(self,
                                   consumer: ConsumerT,
                                   producer: ProducerT,
                                   **kwargs: Any) -> TransactionManagerT:
        ...

    @abc.abstractmethod
    def create_conductor(self, **kwargs: Any) -> ConductorT:
        ...



#!/usr/bin/env python
import asyncio
import aiohttp
import faust


class Record(faust.Record):
    value: int


app = faust.App('concurrency', broker='kafka://localhost')
topic = app.topic('concurrency', value_type=Record)


@app.agent(topic, concurrency=200)
async def mytask(records):
    session = aiohttp.ClientSession()
    async for record in records:
        await session.get(f'http://www.google.com/?#q={record.value}')


async def producer():
    for i in range(10_000):
        await topic.send(value=Record(value=i))


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(producer())
    loop.stop()



"""Utilities for generating code at runtime."""
from typing import Any, Callable, Dict, List, Mapping, Tuple, cast

__all__ = [
    'Function',
    'Method',
    'InitMethod',
    'HashMethod',
    'CompareMethod',
    'EqMethod',
    'NeMethod',
    'LeMethod',
    'LtMethod',
    'GeMethod',
    'GtMethod',
    'build_function',
    'build_function_source',
    'reprkwargs',
    'reprcall',
]

MISSING = object()


def Function(name: str,
             args: List[str],
             body: List[str],
             *,
             globals: Dict[str, Any] = None,
             locals: Dict[str, Any] = None,
             return_type: Any = MISSING,
             argsep: str = ', ') -> Callable:
    """Compile function code object from args and body."""
    return build_function(
        name=name,
        source=build_function_source(
            name=name,
            args=args,
            body=body,
            return_type=return_type,
            argsep=argsep,
        ),
        return_type=return_type,
        globals=globals,
        locals=locals,
    )


def build_closure_source(name: str,
                         args: List[str],
                         body: List[str],
                         *,
                         outer_name: str = '__outer__',
                         outer_args: List[str] = None,
                         closures: Dict[str, str],
                         return_type: Any = MISSING,
                         indentlevel: int = 0,
                         indentspaces: int = 4,
                         argsep: str = ', ') -> str:
    inner_source = build_function_source(
        name, args, body,
        return_type=return_type,
        indentlevel=indentlevel,
        indentspaces=indentspaces,
        argsep=argsep,
    )
    closure_vars = [
        f'{local_name} = {global_name}'
        for local_name, global_name in closures.items()
    ]
    outer_source = build_function_source(
        name=outer_name,
        args=outer_args or [],
        body=closure_vars + inner_source.split('\n') + [f'return {name}'],
        return_type=MISSING,
        indentlevel=indentlevel,
        indentspaces=indentspaces,
        argsep=argsep,
    )
    return outer_source


def build_closure(outer_name: str, source: str, *args: Any,
                  return_type: Any = MISSING,
                  globals: Dict[str, Any] = None,
                  locals: Dict[str, Any] = None) -> Callable:
    assert locals is not None
    if return_type is not MISSING:
        locals['_return_type'] = return_type
    exec(source, globals, locals)
    obj = locals[outer_name](*args)
    obj.__sourcecode__ = source
    return cast(Callable, obj)


def build_function(name: str, source: str,
                   *,
                   return_type: Any = MISSING,
                   globals: Dict[str, Any] = None,
                   locals: Dict[str, Any] = None) -> Callable:
    """Generate function from Python from source code string."""
    assert locals is not None
    if return_type is not MISSING:
        locals['_return_type'] = return_type
    exec(source, globals, locals)
    obj = locals[name]
    obj.__sourcecode__ = source
    return cast(Callable, obj)


def build_function_source(name: str,
                          args: List[str],
                          body: List[str],
                          *,
                          return_type: Any = MISSING,
                          indentlevel: int = 0,
                          indentspaces: int = 4,
                          argsep: str = ', ') -> str:
    """Generate function source code from args and body."""
    indent = ' ' * indentspaces
    curindent = indent * indentlevel
    nextindent = indent * (indentlevel + 1)
    return_annotation = ''
    if return_type is not MISSING:
        return_annotation = '->_return_type'
    bodys = '\n'.join(f'{nextindent}{b}' for b in body)
    return (f'{curindent}def {name}({argsep.join(args)}){return_annotation}:\n'
            f'{bodys}')


def Method(name: str,
           args: List[str],
           body: List[str],
           **kwargs: Any) -> Callable:
    """Generate Python method."""
    return Function(name, ['self'] + args, body, **kwargs)


def InitMethod(args: List[str],
               body: List[str],
               **kwargs: Any) -> Callable[[], None]:
    """Generate ``__init__`` method."""
    return Method('__init__', args, body, return_type='None', **kwargs)


def HashMethod(attrs: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__hash__`` method."""
    self_tuple = obj_attrs_tuple('self', attrs)
    return Method('__hash__',
                  [],
                  [f'return hash({self_tuple})'],
                  **kwargs)


def EqMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__eq__`` method."""
    return CompareMethod(name='__eq__', op='==', fields=fields, **kwargs)


def NeMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__ne__`` method."""
    return CompareMethod(name='__ne__', op='!=', fields=fields, **kwargs)


def GeMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__ge__`` method."""
    return CompareMethod(name='__ge__', op='>=', fields=fields, **kwargs)


def GtMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__gt__`` method."""
    return CompareMethod(name='__gt__', op='>', fields=fields, **kwargs)


def LeMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__le__`` method."""
    return CompareMethod(name='__le__', op='<=', fields=fields, **kwargs)


def LtMethod(fields: List[str], **kwargs: Any) -> Callable[[], None]:
    """Generate ``__lt__`` method."""
    return CompareMethod(name='__lt__', op='<', fields=fields, **kwargs)


def CompareMethod(name: str,
                  op: str,
                  fields: List[str],
                  **kwargs: Any) -> Callable[[], None]:
    """Generate object comparison method.

    Excellent for ``__eq__``, ``__le__``, etc.

    Examples:
        The example:

        .. sourcecode:: python

            CompareMethod(
                name='__eq__',
                op='==',
                fields=['x', 'y'],
            )

        Generates a method like this:

        .. sourcecode:: python

           def __eq__(self, other):
              if other.__class__ is self.__class__:
                   return (self.x,self.y) == (other.x,other.y)
               return NotImplemented
    """
    self_tuple = obj_attrs_tuple('self', fields)
    other_tuple = obj_attrs_tuple('other', fields)
    return Method(name,
                  ['other'],
                  ['if other.__class__ is self.__class__:',
                   f' return {self_tuple}{op}{other_tuple}',
                   'return NotImplemented'],
                  **kwargs)


def obj_attrs_tuple(obj_name: str, attrs: List[str]) -> str:
    """Draw Python tuple from list of attributes.

    If attrs is ``['x', 'y']`` and ``obj_name`` is 'self',
    returns ``(self.x,self.y)``.
    """
    if not attrs:
        return '()'
    return f'({",".join([f"{obj_name}.{f}" for f in attrs])},)'


def reprkwargs(kwargs: Mapping[str, Any], *,
               sep: str = ', ',
               fmt: str = '{0}={1}') -> str:
    return sep.join(fmt.format(k, repr(v)) for k, v in kwargs.items())


def reprcall(name: str,
             args: Tuple = (),
             kwargs: Mapping[str, Any] = {},  # noqa: B006
             *,
             sep: str = ', ') -> str:
    return '{0}({1}{2}{3})'.format(
        name, sep.join(map(repr, args or ())),
        (args and kwargs) and sep or '',
        reprkwargs(kwargs, sep=sep),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import re
import sys
from setuptools import Extension, find_packages, setup
from distutils.command.build_ext import build_ext
from distutils.errors import (
    CCompilerError,
    DistutilsExecError,
    DistutilsPlatformError,
)

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = os.environ.get('USE_CYTHON', True)


if os.environ.get('NO_CYTHON'):
    USE_CYTHON = False

NAME = 'faust'
BUNDLES = {
    'aiodns',
    'aiomonitor',
    'cchardet',
    'ckafka',
    'ciso8601',
    'cython',
    'datadog',
    'debug',
    'fast',
    'orjson',
    'redis',
    'rocksdb',
    'rocksdict',
    'setproctitle',
    'statsd',
    'uvloop',
    'eventlet',
    'yaml',
    'prometheus',
}
CFLAGS = ['-O2']
LDFLAGS = []
LIBRARIES = []
E_UNSUPPORTED_PYTHON = NAME + ' 1.0 requires Python %%s or later!'

if sys.version_info < (3, 7):
    raise Exception(E_UNSUPPORTED_PYTHON % ('3.7',))  # NOQA

from pathlib import Path  # noqa

README = Path('README.rst')

# -*- Compiler Flags -*-

if sys.platform == 'win32':
    LDFLAGS.append('ws2_32.lib')
else:
    CFLAGS.extend(['-Wall', '-Wsign-compare', '-Wconversion'])
    LIBRARIES.append('z')

# -*- C Extensions -*-
ext = '.pyx' if USE_CYTHON else '.c'

extensions = [
    Extension(
        'faust._cython.windows',
        ['faust/_cython/windows' + ext],
        libraries=LIBRARIES,
        extra_compile_args=CFLAGS,
        extra_link_args=LDFLAGS,
    ),
    Extension(
        'faust._cython.streams',
        ['faust/_cython/streams' + ext],
        libraries=LIBRARIES,
        extra_compile_args=CFLAGS,
        extra_link_args=LDFLAGS,
    ),
    Extension(
        'faust.transport._cython.conductor',
        ['faust/transport/_cython/conductor' + ext],
        libraries=LIBRARIES,
        extra_compile_args=CFLAGS,
        extra_link_args=LDFLAGS,
    ),
]


if USE_CYTHON:
    print('---*--- USING CYTHON ---*---')
    extensions = cythonize(extensions)


class BuildFailed(Exception):
    pass


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError,
                DistutilsPlatformError, ValueError):
            raise BuildFailed()


# -*- Distribution Meta -*-

re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_doc = re.compile(r'^"""(.+?)"""')


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, attr_value.strip("\"'")),)


def add_doc(m):
    return (('doc', m.groups()[0]),)


pats = {re_meta: add_default, re_doc: add_doc}
here = Path(__file__).parent.absolute()
with open(here / NAME / '__init__.py') as meta_fh:
    meta = {}
    for line in meta_fh:
        if line.strip() == '# -eof meta-':
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))

# -*- Installation Requires -*-


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*root, *path.split('/'))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / 'requirements').joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs('extras', *p)


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + '.txt') for x in BUNDLES}


# -*- Long Description -*-


if README.exists():
    long_description = README.read_text(encoding='utf-8')
else:
    long_description = 'See http://pypi.org/project/{}'.format(NAME)

# -*- %%% -*-


def do_setup(**kwargs):
    setup(
        name='twm-faust',
        version=meta['version'],
        description=meta['doc'],
        long_description=long_description,
        long_description_content_type='text/x-rst',
        author=meta['author'],
        author_email=meta['contact'],
        url=meta['homepage'],
        platforms=['any'],
        license='BSD 3-Clause',
        packages=find_packages(exclude=['examples', 'ez_setup', 't', 't.*']),
        # PEP-561: https://www.python.org/dev/peps/pep-0561/
        package_data={'faust': ['py.typed']},
        include_package_data=True,
        python_requires='>=3.6.0',
        zip_safe=False,
        install_requires=reqs('default.txt'),
        tests_require=reqs('test.txt'),
        extras_require=extras_require(),
        entry_points={
            'console_scripts': [
                'faust = faust.cli.faust:cli',
            ],
        },
        project_urls={
            'Bug Reports': 'https://github.com/robinhood/faust/issues',
            'Source': 'https://github.com/robinhood/faust',
            'Documentation': 'https://faust.readthedocs.io/',
        },
        keywords=[
            'stream',
            'processing',
            'asyncio',
            'distributed',
            'queue',
            'kafka',
        ],
        classifiers=[
            'Framework :: AsyncIO',
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Natural Language :: English',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Operating System :: POSIX',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX :: BSD',
            'Operating System :: Microsoft :: Windows',
            'Topic :: System :: Networking',
            'Topic :: System :: Distributed Computing',
        ],
        **kwargs)


try:
    do_setup(cmdclass={'build_ext': ve_build_ext},
             ext_modules=extensions)
except BuildFailed:
    print('************************************************************')
    print('Cannot compile C accelerated modules, using pure python')
    print('************************************************************')
    do_setup()



from types import SimpleNamespace
import pytest
from mode.utils.mocks import Mock
from faust.livecheck.locals import current_test_stack
from faust.livecheck.patches.aiohttp import (
    LiveCheckMiddleware,
    patch_aiohttp_session,
)


@pytest.mark.asyncio
async def test_patch_aiohttp_session(*, execution):

    patch_aiohttp_session()
    from aiohttp.client import ClientSession
    trace_configs = []
    async with ClientSession(trace_configs=trace_configs):
        assert trace_configs
        config = trace_configs[-1]
        on_request_start = config.on_request_start[-1]

        session = Mock()
        ctx = SimpleNamespace()
        params = Mock()
        params.headers = {}

        assert current_test_stack.top is None

        await on_request_start(session, ctx, params)

        with current_test_stack.push(execution):
            await on_request_start(session, ctx, params)
            assert params.headers
            assert 'LiveCheck-Test-Id' in params.headers

    async with ClientSession(trace_configs=None):
        pass


@pytest.mark.asyncio
async def test_LiveCheckMiddleware(*, execution):
    request = Mock()
    request.headers = execution.as_headers()
    assert current_test_stack.top is None

    called = [False]

    async def handler(request):
        called[0] = True
        assert current_test_stack.top is not None

    await LiveCheckMiddleware()(request, handler)

    assert current_test_stack.top is None
    assert called[0]


@pytest.mark.asyncio
async def test_LiveCheckMiddleware__no_test(*, execution):
    request = Mock()
    request.headers = {}
    assert current_test_stack.top is None

    called = [False]

    async def handler(request):
        called[0] = True
        assert current_test_stack.top is None

    await LiveCheckMiddleware()(request, handler)

    assert current_test_stack.top is None
    assert called[0]



"""
Celery-like task queue implemented using Faust.

This example comes with a background thread that
sends a task every second.

After starting Kafka locally, you can run the example:

.. sourcecode:: console

    $ python examples/task_queue.py worker -l info

You can also see stats about the worker by visiting http://localhost:6066.
"""
import random
from typing import Any, Awaitable, Callable, Mapping, MutableMapping, Sequence
import faust
from mode.utils.objects import qualname


class Request(faust.Record):
    """Describes how tasks are serialized and sent to Kafka."""

    #: Correlation ID, can be used to pass results back to caller.
    id: str

    #: Name of the task as registered in the task_registry.
    name: str

    #: Positional arguments to the task.
    arguments: Sequence

    #: Keyword arguments to the task.
    keyword_arguments: Mapping

    async def __call__(self) -> Any:
        return await self.handler(*self.arguments, **self.keyword_arguments)

    @property
    def handler(self) -> Callable[..., Awaitable]:
        return task_registry[self.name]


app = faust.App('faust-celery', broker='kafka://localhost')

task_queue_topic = app.topic('tasks', value_type=Request)

task_registry: MutableMapping[str, Callable[..., Awaitable]]
task_registry = {}


@app.agent(task_queue_topic)
async def process_task(tasks: faust.Stream[Request]) -> None:
    """A "worker" stream processor that executes tasks."""
    async for task in tasks:
        print(f'Processing task: {task!r}')
        result = await task()
        print(f'Result of {task.id} is: {result!r}')


class Task:

    def __init__(self, fun: Callable[..., Awaitable],
                 *,
                 name: str = None) -> None:
        self.fun: Callable[..., Awaitable] = fun
        self.name = name or qualname(fun)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.fun(*args, **kwargs)

    async def delay(self, *args: Any, **kwargs: Any) -> Any:
        return await self.apply_async(args, kwargs)

    async def apply_async(self,
                          args: Sequence,
                          kwargs: Mapping,
                          id: str = None,
                          **options: Any) -> None:
        id = id or faust.uuid()
        return await process_task.send(value=Request(
            id=id,
            name=self.name,
            arguments=args,
            keyword_arguments=kwargs,
        ))


def task(fun: Callable) -> Task:
    # Task decorator
    task = Task(fun)
    task_registry[task.name] = task
    return task


@task
async def add(x: int, y: int) -> int:
    return x + y


@app.timer(1.0)
async def _send_tasks() -> None:
    await add.delay(random.randint(0, 100), random.randint(0, 100))


if __name__ == '__main__':
    app.main()



import asyncio
import random
import sys
from typing import Any, Set
from faust.cli import option
from mode.utils.aiter import aiter, anext


def install_produce_command(app) -> None:
    @app.command(
        option('--max-latency',
               type=float, default=0.5, envvar='PRODUCE_LATENCY',
               help='Add delay of (at most) n seconds between publishing.'),
        option('--max-messages',
               type=int, default=None,
               help='Send at most N messages or 0 for infinity.'),
    )
    async def produce(self, max_latency: float, max_messages: int):
        """Produce example events."""
        prods = {aiter(p(max_messages)) for p in app.stress_producers}
        i = 0
        while not app.should_stop:
            to_remove: Set[Any] = set()
            for producer in prods:
                i += 1
                try:
                    await anext(producer)
                except StopAsyncIteration:
                    to_remove.add(producer)
                if not max_latency:
                    # no latency, print every 10,000 messages
                    if not i % 10000:
                        self.say(f'+SEND {i}')
                else:
                    # with latency, print every 10 messages
                    if not i % 10:
                        self.say(f'+SEND {i}')
            if not prods:
                await asyncio.sleep(1.0)
            if max_latency:
                await asyncio.sleep(random.uniform(0, max_latency))
            for producer in to_remove:
                prods.discard(producer)
        print('No more producers - exiting', file=sys.stderr)



import socket
import faust
from . import states
from .models import Error, Status

REPORT_TOPIC_NAME = 'f-systemcheck'
ERRORS_TOPIC_NAME = 'f-errors'


def get_reporting_topic(app):
    return app.topic(REPORT_TOPIC_NAME, value_type=Status)


def get_error_topic(app):
    return app.topic(ERRORS_TOPIC_NAME, value_type=Error)


async def send_update(app, status: Status):
    return await get_reporting_topic(app).send(value=status)


async def send_simple_good_status(app, category: str,
                                  state: str = states.OK,
                                  color: str = 'green',
                                  count: int = 1,
                                  severity: str = 'INFO') -> None:
    status = Status(
        app_id=app.conf.id,
        hostname=socket.gethostname(),
        category=category,
        state=state,
        color=color,
        count=count,
        severity=severity,

    )
    await send_update(app, status)


async def send_simple_bad_status(app, category: str,
                                 state: str = states.STALL,
                                 color: str = 'red',
                                 count: int = 1,
                                 severity: str = 'CRIT') -> None:
    status = Status(
        app_id=app.conf.id,
        hostname=socket.gethostname(),
        category=category,
        state=state,
        color=color,
        count=count,
        severity=severity,

    )
    await send_update(app, status)


class SimpleCheck:

    def __init__(self, category: str) -> None:
        self.category: str = category
        self.failed: int = 0

    async def send_ok(self, app) -> None:
        self.failed = 0
        await send_simple_good_status(app, self.category)

    async def send_fail(self, app) -> None:
        self.failed += 1
        await send_simple_bad_status(app, self.category, count=self.failed)


class DashboardApp(faust.App):

    def on_webserver_init(self, web):
        from . import assets
        web.add_static('/dashboard/assets/', assets.get_path())


def get_reporting_app() -> DashboardApp:
    from ..app import create_app
    return create_app(
        'f-stress-systemcheck',
        base=DashboardApp,
        origin='t.stress.reports',
    )



import pytest
from mode.utils.mocks import AsyncMock, Mock
from faust import Event


class test_Event:

    @pytest.fixture
    def key(self):
        return Mock(name='key')

    @pytest.fixture
    def value(self):
        return Mock(name='value')

    @pytest.fixture
    def message(self):
        return Mock(name='message')

    @pytest.fixture
    def event(self, *, app, key, value, message):
        return Event(app, key, value, {}, message)

    @pytest.mark.asyncio
    async def test_send(self, *, event):
        callback = Mock(name='callback')
        schema = Mock(name='schema')
        event._send = AsyncMock(name='event._send')
        await event.send(
            channel='chan',
            key=event.key,
            value=event.value,
            partition=3,
            timestamp=None,
            headers={'k': 'v'},
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        event._send.assert_called_once_with(
            'chan',
            event.key,
            event.value,
            3,
            None,
            {'k': 'v'},
            schema,
            'kser',
            'vser',
            callback,
            force=False,
        )

    @pytest.mark.asyncio
    async def test_send__USE_EXISTING_KEY_VALUE(self, *, event):
        callback = Mock(name='callback')
        event._send = AsyncMock(name='event._send')
        event.headers = {'k': 'v'}
        await event.send(
            channel='chan',
            partition=3,
            timestamp=None,
            schema=None,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        event._send.assert_called_once_with(
            'chan',
            event.key,
            event.value,
            3,
            None,
            event.headers,
            None,
            'kser',
            'vser',
            callback,
            force=False,
        )

    @pytest.mark.asyncio
    async def test_forward(self, *, event):
        callback = Mock(name='callback')
        event._send = AsyncMock(name='event._send')
        await event.forward(
            channel='chan',
            key=event.key,
            value=event.value,
            headers={'k': 'v'},
            partition=3,
            timestamp=1234,
            schema=None,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        event._send.assert_called_once_with(
            'chan',
            event.key,
            event.value,
            3,
            1234,
            {'k': 'v'},
            None,
            'kser',
            'vser',
            callback,
            force=False,
        )

    @pytest.mark.asyncio
    async def test_forward__USE_EXISTING_KEY_VALUE(self, *, event):
        callback = Mock(name='callback')
        event._send = AsyncMock(name='event._send')
        event.message.headers = {'k1': 'v1'}
        schema = Mock(name='schema')
        await event.forward(
            channel='chan',
            partition=3,
            timestamp=None,
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        event._send.assert_called_once_with(
            'chan',
            event.message.key,
            event.message.value,
            3,
            None,
            event.message.headers,
            schema,
            'kser',
            'vser',
            callback,
            force=False,
        )

    def test_attach(self, *, event, app):
        callback = Mock(name='callback')
        schema = Mock(name='schema')
        app._attachments.put = Mock(name='_attachments.put')
        result = event._attach(
            channel='chan',
            key=b'k',
            value=b'v',
            partition=3,
            timestamp=None,
            headers={'k': 'v'},
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        app._attachments.put.assert_called_once_with(
            event.message,
            'chan',
            b'k',
            b'v',
            partition=3,
            timestamp=None,
            headers={'k': 'v'},
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
        )
        assert result is app._attachments.put()

    @pytest.mark.asyncio
    async def test__send(self, *, event, app):
        app._attachments.maybe_put = AsyncMock(name='maybe_put')
        callback = Mock(name='callback')
        schema = Mock(name='schema')
        await event._send(
            channel='chan',
            key=b'k',
            value=b'v',
            partition=4,
            timestamp=33.31234,
            headers=[('k', 'v')],
            schema=schema,
            key_serializer='kser',
            value_serializer='vser',
            callback=callback,
            force=True,
        )

        app._attachments.maybe_put.assert_called_once_with(
            'chan',
            b'k',
            b'v',
            4,
            33.31234,
            [('k', 'v')],
            schema,
            'kser',
            'vser',
            callback,
            force=True,
        )

    def test_repr(self, *, event):
        assert repr(event)

    @pytest.mark.asyncio
    async def test_AsyncContextManager(self, *, event):
        event.ack = Mock(name='ack')
        block_executed = False
        async with event:
            block_executed = True
        assert block_executed

        event.ack.assert_called_once_with()



#!/usr/bin/env python
import re
from pathlib import Path
from setuptools import find_packages, setup

NAME = 'proj'

README = Path('README.rst')

# -*- Classifiers -*-

classes = '''
    Development Status :: 4 - Beta
    License :: OSI Approved :: BSD License
    Programming Language :: Python
    Programming Language :: Python :: 3 :: Only
    Programming Language :: Python :: 3.6
    Operating System :: POSIX
    Operating System :: POSIX :: Linux
    Operating System :: MacOS :: MacOS X
    Operating System :: POSIX :: BSD
    Operating System :: Microsoft :: Windows
    Topic :: Framework :: Django
'''
classifiers = [s.strip() for s in classes.split('\n') if s]

# -*- Distribution Meta -*-

re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_doc = re.compile(r'^"""(.+?)"""')


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, attr_value.strip("\"'")),)


def add_doc(m):
    return (('doc', m.groups()[0]),)


pats = {re_meta: add_default, re_doc: add_doc}
here = Path(__file__).parent.absolute()
with open(here / NAME / '__init__.py') as meta_fh:
    meta = {}
    for line in meta_fh:
        if line.strip() == '# -eof meta-':
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))

# -*- Installation Requires -*-


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def _pip_requirement(req):
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*path.split('/'))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / 'requirements').joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]

# -*- Long Description -*-


if README.exists():
    long_description = README.read_text(encoding='utf-8')
else:
    long_description = 'See http://pypi.org/project/{}'.format(NAME)

# -*- %%% -*-

setup(
    name=NAME,
    version=meta['version'],
    description=meta['doc'],
    author=meta['author'],
    author_email=meta['contact'],
    url=meta['homepage'],
    platforms=['any'],
    license='BSD',
    packages=find_packages(exclude=['ez_setup', 'tests', 'tests.*']),
    include_package_data=True,
    python_requires='>=3.6.0',
    keywords=[],
    zip_safe=False,
    install_requires=reqs('default.txt'),
    tests_require=reqs('test.txt'),
    classifiers=classifiers,
    long_description=long_description,
    entry_points={
        'console_scripts': [
            'proj = proj.__main__:main',
            'proj-faust = faustapp.app:main',
        ],
    },
)



"""Faust Web."""
from .base import Request, Response, Web
from .blueprints import Blueprint
from .views import View, gives_model, takes_model
from faust.types.web import ResourceOptions

__all__ = [
    'Request',
    'Response',
    'ResourceOptions',
    'Web',
    'Blueprint',
    'View',
    'gives_model',
    'takes_model',
]



class test_Arena:

    def test_json(self, faust_json):
        exitcode, model, stderr = faust_json('model', 'app.Arena')
        assert not exitcode

        assert model == [
            {'field': 'points',
             'type': 'typing.List[__main__.Point]',
             'default': '*'},
            {'field': 'timestamp',
             'type': 'float',
             'default': 'None'},
        ]

    def test_tabulated(self, faust):
        exitcode, stdout, stderr = faust('model', 'app.Arena')
        assert not exitcode
        assert b'typing.List' in stdout

    def test_colors(self, faust_color):
        exitcode, stdout, stderr = faust_color('model', 'app.Arena')
        assert b'typing.List' in stdout


class test_Point:

    def test_json(self, faust_json):
        exitcode, model, stderr = faust_json('model', 'app.Point')
        assert not exitcode

        assert model == [
            {'field': 'x', 'type': 'int', 'default': '*'},
            {'field': 'y', 'type': 'int', 'default': '*'},
        ]

    def test_tabulated(self, faust):
        exitcode, stdout, stderr = faust('model', 'app.Point')
        assert not exitcode
        assert b'int' in stdout

    def test_colors(self, faust_color):
        exitcode, stdout, stderr = faust_color('model', 'app.Point')
        assert not exitcode
        assert b'int' in stdout



import logging
import os
from typing import List
from raven import Client
from raven.handlers.logging import SentryHandler
from raven_aiohttp import AioHttpTransport

__all__ = [
    'broker',
    'store',
    'topic_partitions',
    'sentry_dsn',
    'loghandlers',
]

broker: str = os.environ.get('STRESS_BROKER', 'kafka://127.0.0.1:9092')
store: str = os.environ.get('STRESS_STORE', 'rocksdb://')
topic_partitions: int = int(os.environ.get('STRESS_PARTITIONS', 4))
sentry_dsn: str = os.environ.get('SENTRY_DSN')


def loghandlers() -> List[SentryHandler]:
    handlers = []
    if sentry_dsn:
        client = Client(
            dsn=sentry_dsn,
            include_paths=[__name__.split('.', 1)[0]],
            transport=AioHttpTransport,
            disable_existing_loggers=False,
        )
        handler = SentryHandler(client)
        handler.setLevel(logging.ERROR)
        handler.propagate = False
        handlers.append(handler)
    return handlers



"""Window Types."""
import os
import typing

from math import floor
from typing import List, Type, cast
from mode import Seconds, want_seconds

from .types.windows import WindowRange, WindowRange_from_start, WindowT

__all__ = [
    'Window',
    'HoppingWindow',
    'TumblingWindow',
    'SlidingWindow',
]

NO_CYTHON = bool(os.environ.get('NO_CYTHON', False))


class Window(WindowT):
    """Base class for window types."""


class _PyHoppingWindow(Window):
    """Hopping window type.

    Fixed-size, overlapping windows.
    """

    size: float
    step: float

    def __init__(self, size: Seconds, step: Seconds,
                 expires: Seconds = None) -> None:
        self.size = want_seconds(size)
        self.step = want_seconds(step)
        self.expires = want_seconds(expires) if expires else None

    def ranges(self, timestamp: float) -> List[WindowRange]:
        start = self._start_initial_range(timestamp)
        return [
            WindowRange_from_start(float(start), self.size)
            for start in range(int(start), int(timestamp) + 1, int(self.step))
        ]

    def stale(self, timestamp: float, latest_timestamp: float) -> bool:
        return (timestamp <= self._stale_before(latest_timestamp, self.expires)
                if self.expires else False)

    def current(self, timestamp: float) -> WindowRange:
        """Get the latest window range for a given timestamp."""
        step = self.step
        start = self._start_initial_range(timestamp)
        m = floor((timestamp - start) / step)
        return WindowRange_from_start(start + (step * m), self.size)

    def delta(self, timestamp: float, d: Seconds) -> WindowRange:
        return self.current(timestamp - want_seconds(d))

    def earliest(self, timestamp: float) -> WindowRange:
        start = self._start_initial_range(timestamp)
        return WindowRange_from_start(float(start), self.size)

    def _start_initial_range(self, timestamp: float) -> float:
        closest_step = (timestamp // self.step) * self.step
        return closest_step - self.size + self.step

    def _stale_before(self, latest_timestamp: float, expires: float) -> float:
        return self.current(latest_timestamp - expires)[0]


if typing.TYPE_CHECKING:
    HoppingWindow = _PyHoppingWindow
else:
    if not NO_CYTHON:  # pragma: no cover
        try:
            from ._cython.windows import HoppingWindow
        except ImportError:
            HoppingWindow = _PyHoppingWindow
        else:
            # isinstance(HoppingWindow, Window) is True
            # isinstance(HoppingWindow, WindowT) is True
            Window.register(HoppingWindow)
    else:  # pragma: no cover
        HoppingWindow = _PyHoppingWindow


class TumblingWindow(HoppingWindow):
    """Tumbling window type.

    Fixed-size, non-overlapping, gap-less windows.
    """

    def __init__(self, size: Seconds, expires: Seconds = None) -> None:
        super(TumblingWindow, self).__init__(size, size, expires)


class _PySlidingWindow(Window):
    """Sliding window type.

    Notes:
        Fixed-size, overlapping windows that work on differences between
        record timestamps
    """

    before: float
    after: float

    def __init__(self, before: Seconds, after: Seconds,
                 expires: Seconds) -> None:
        self.before = want_seconds(before)
        self.after = want_seconds(after)
        self.expires = want_seconds(expires)

    def ranges(self, timestamp: float) -> List[WindowRange]:
        """Return list of windows from timestamp.

        Notes:
            .. sourcecode:: sql

                SELECT *
                  FROM s1, s2
                WHERE
                    s1.key = s2.key
                AND
                s1.ts - before <= s2.ts AND s2.ts <= s1.ts + after

        """
        return [
            (timestamp - self.before, timestamp + self.after),
        ]

    def stale(self, timestamp: float, latest_timestamp: float) -> bool:
        return (timestamp <= self._stale_before(self.expires, latest_timestamp)
                if self.expires else False)

    def _stale_before(self, expires: float, latest_timestamp: float) -> float:
        return latest_timestamp - expires

    def current(self, timestamp: float) -> WindowRange:
        """Get the latest window range for a given timestamp."""
        return timestamp - self.before, timestamp + self.after

    def delta(self, timestamp: float, d: Seconds) -> WindowRange:
        return self.current(timestamp - want_seconds(d))

    def earliest(self, timestamp: float) -> WindowRange:
        return self.current(timestamp)


if not NO_CYTHON:  # pragma: no cover
    try:
        from ._cython.windows import SlidingWindow
    except ImportError:
        SlidingWindow = _PySlidingWindow
    else:
        SlidingWindow = cast(Type[Window], SlidingWindow)
        Window.register(SlidingWindow)
else:  # pragma: no cover
    SlidingWindow = _PySlidingWindow



import logging
from faust.utils.terminal.spinners import Spinner, SpinnerHandler
from mode.utils.mocks import Mock, call, patch


def spinner(file=None, isatty=True):
    file = file if file is not None else Mock(name='file')
    file.isatty.return_value = isatty
    return Spinner(file)


class test_Spinner:

    def spinner(self, file=None, isatty=True):
        return spinner(file=file, isatty=isatty)

    def test_constructor(self):
        f = Mock(name='file')
        spinner = self.spinner(f)
        assert spinner.file is f
        assert spinner.width == 0
        assert spinner.count == 0
        assert not spinner.stopped

    def test_update(self):
        s = self.spinner()
        assert s.count == 0
        s.begin = Mock(name='begin')
        s.write = Mock(name='write')
        s.update()
        s.begin.assert_called_once_with()
        assert s.count == 1
        s.write.assert_called_once_with(s.sprites[0])

        for i in range(100):
            s.update()
            s.begin.assert_called_once_with()
            assert s.count == i + 2
            s.write.assert_called_with(
                s.sprites[(s.count - 1) % len(s.sprites)])
        count = s.count
        s.stop()
        assert s.stopped
        s.update()
        assert s.count == count

    def test_reset(self):
        s = self.spinner(isatty=True)
        s.reset()
        assert not s.stopped
        assert s.count == 0

    def test_write(self):
        s = self.spinner(isatty=True)
        s.write('f')
        s.file.write.assert_has_calls([
            call('f'), call(''),
        ])
        s.file.flush.assert_called_once_with()

    def test_write__notatty(self):
        s = self.spinner(isatty=False)
        s.write('f')
        s.file.write.assert_not_called()

    def test_begin(self):
        s = self.spinner()
        with patch('atexit.register') as atexit_register:
            s.begin()
            s.file.write.assert_has_calls([
                call(s.cursor_hide), call(''),
            ])
            s.file.flush.assert_called_once_with()
            atexit_register.assert_called_with(
                type(s)._finish, s.file, at_exit=True)

    def test_finish(self):
        s = self.spinner()
        s.finish()
        s.file.write.assert_has_calls([
            call(s.bell), call(''), call(s.cursor_show), call(''),
        ])
        s.file.flush.assert_called_once_with()
        assert s.stopped


def test_SpinnerHandler():
    s = Mock(name='spinner', autospec=Spinner)
    s.stopped = False
    handler = SpinnerHandler(spinner=s)
    assert handler.spinner is s
    handler.emit(Mock(name='logrecord', autospec=logging.LogRecord))
    handler.spinner.update.assert_called_once_with()


def test_SpinnerHandler__no_spinner():
    SpinnerHandler(spinner=None).emit(
        Mock(
            name='logrecord',
            autospec=logging.LogRecord,
        ),
    )



"""Channel.

A channel is used to send values to streams.

The stream will iterate over incoming events in the channel.

"""
import asyncio

from typing import (
    Any,
    Awaitable,
    Callable,
    Mapping,
    MutableSet,
    Optional,
    Set,
    Tuple,
    TypeVar,
    cast,
    no_type_check,
)
from weakref import WeakSet

from mode import Seconds, get_logger, want_seconds
from mode.utils.futures import maybe_async, stampede
from mode.utils.queues import ThrowableQueue

from .types import (
    AppT,
    ChannelT,
    CodecArg,
    EventT,
    FutureMessage,
    K,
    Message,
    MessageSentCallback,
    ModelArg,
    PendingMessage,
    RecordMetadata,
    SchemaT,
    StreamT,
    TP,
    V,
)
from .types.core import HeadersArg, OpenHeadersArg, prepare_headers
from .types.tuples import _PendingMessage_to_Message

__all__ = ['Channel']

logger = get_logger(__name__)

T = TypeVar('T')
T_contra = TypeVar('T_contra', contravariant=True)


class Channel(ChannelT[T]):
    """Create new channel.

    Arguments:
        app: The app that created this channel (``app.channel()``)

        schema: Schema used for serialization/deserialization
        key_type:  The Model used for keys in this channel.
           (overrides schema if one is defined)
        value_type: The Model used for values in this channel.
           (overrides schema if one is defined)
        maxsize: The maximum number of messages this channel can hold.
                 If exceeded any new ``put`` call will block until a message
                 is removed from the channel.
        is_iterator: When streams iterate over a channel they will call
            ``stream.clone(is_iterator=True)`` so this attribute
            denotes that this channel instance is currently being iterated
            over.
        active_partitions: Set of active topic partitions this
           channel instance is assigned to.
        loop: The :mod:`asyncio` event loop to use.
    """

    app: AppT
    schema: SchemaT
    key_type: Optional[ModelArg]
    value_type: Optional[ModelArg]
    is_iterator: bool

    _queue: Optional[ThrowableQueue]
    _root: Optional['Channel']
    _subscribers: MutableSet['Channel']

    def __init__(self,
                 app: AppT,
                 *,
                 schema: SchemaT = None,
                 key_type: ModelArg = None,
                 value_type: ModelArg = None,
                 is_iterator: bool = False,
                 queue: ThrowableQueue = None,
                 maxsize: int = None,
                 root: ChannelT = None,
                 active_partitions: Set[TP] = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        self.app = app
        self.loop = loop
        self.is_iterator = is_iterator
        self._queue = queue
        self.maxsize = maxsize
        self.deliver = self._compile_deliver()  # type: ignore
        self._root = cast(Channel, root)
        self.active_partitions = active_partitions
        self._subscribers = WeakSet()

        if schema is None:
            self.schema = self._get_default_schema(key_type, value_type)
        else:
            self.schema = schema
            self.schema.update(key_type=key_type, value_type=value_type)
        self.key_type = self.schema.key_type
        self.value_type = self.schema.value_type

    def _get_default_schema(self,
                            key_type: ModelArg = None,
                            value_type: ModelArg = None) -> SchemaT:
        return self.app.conf.Schema(
            key_type=key_type,
            value_type=value_type,
        )

    @property
    def queue(self) -> ThrowableQueue:
        """Return the underlying queue/buffer backing this channel."""
        if self._queue is None:
            # this should only be set after clone = channel.__aiter__()
            # which means the loop is not accessed by merely defining
            # a channel at module scope.
            maxsize = self.maxsize
            if maxsize is None:
                maxsize = self.app.conf.stream_buffer_maxsize
            self._queue = self.app.FlowControlQueue(
                maxsize=maxsize,
                loop=self.loop,
                clear_on_resume=True,
            )
        return self._queue

    def clone(self, *, is_iterator: bool = None, **kwargs: Any) -> ChannelT[T]:
        """Create clone of this channel.

        Arguments:
            is_iterator: Set to True if this is now a channel
                that is being iterated over.

        Keyword Arguments:
            **kwargs: Any keyword arguments passed will override any
                of the arguments supported by
                :class:`Channel.__init__ <Channel>`.
        """
        is_it = is_iterator if is_iterator is not None else self.is_iterator
        subchannel: ChannelT = self._clone(is_iterator=is_it, **kwargs)
        if is_it:
            (self._root or self)._subscribers.add(cast(Channel, subchannel))
        # make sure queue is created at this point
        # ^ it's a cached_property
        subchannel.queue
        return subchannel

    def clone_using_queue(self, queue: asyncio.Queue) -> ChannelT[T]:
        """Create clone of this channel using specific queue instance."""
        return self.clone(queue=queue, is_iterator=True)

    def _clone(self, **kwargs: Any) -> ChannelT[T]:
        return type(self)(**{**self._clone_args(), **kwargs})

    def _clone_args(self) -> Mapping:
        # How to create a copy of this channel.
        return {
            'app': self.app,
            'loop': self.loop,
            'schema': self.schema,
            'key_type': self.key_type,
            'value_type': self.value_type,
            'maxsize': self.maxsize,
            'root': self._root if self._root is not None else self,
            'queue': None,
            'active_partitions': self.active_partitions,
        }

    def stream(self, **kwargs: Any) -> StreamT[T]:
        """Create stream reading from this channel."""
        return self.app.stream(self, **kwargs)

    def get_topic_name(self) -> str:
        """Get the topic name, or raise if this is not a named channel."""
        raise NotImplementedError('Channels are unnamed topics')

    async def send(self,
                   *,
                   key: K = None,
                   value: V = None,
                   partition: int = None,
                   timestamp: float = None,
                   headers: HeadersArg = None,
                   schema: SchemaT = None,
                   key_serializer: CodecArg = None,
                   value_serializer: CodecArg = None,
                   callback: MessageSentCallback = None,
                   force: bool = False) -> Awaitable[RecordMetadata]:
        """Send message to channel."""
        return await self._send_now(
            key,
            value,
            partition=partition,
            timestamp=timestamp,
            headers=headers,
            schema=schema,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            callback=callback,
        )

    def send_soon(self,
                  *,
                  key: K = None,
                  value: V = None,
                  partition: int = None,
                  timestamp: float = None,
                  headers: HeadersArg = None,
                  schema: SchemaT = None,
                  key_serializer: CodecArg = None,
                  value_serializer: CodecArg = None,
                  callback: MessageSentCallback = None,
                  force: bool = False,
                  eager_partitioning: bool = False) -> FutureMessage:
        """Produce message by adding to buffer.

        This method is only supported by :class:`~faust.Topic`.

        Raises:
            NotImplementedError: always for in-memory channel.
        """
        raise NotImplementedError()

    def as_future_message(
            self,
            key: K = None,
            value: V = None,
            partition: int = None,
            timestamp: float = None,
            headers: HeadersArg = None,
            schema: SchemaT = None,
            key_serializer: CodecArg = None,
            value_serializer: CodecArg = None,
            callback: MessageSentCallback = None,
            eager_partitioning: bool = False) -> FutureMessage:
        """Create promise that message will be transmitted."""
        open_headers = self.prepare_headers(headers)
        final_key, open_headers = self.prepare_key(
            key, key_serializer, schema, open_headers)
        final_value, open_headers = self.prepare_value(
            value, value_serializer, schema, open_headers)
        if partition is None and eager_partitioning:
            # Note: raises NotImplementedError if used on unnamed channel.
            partition = self.app.producer.key_partition(
                self.get_topic_name(), final_key).partition
        return FutureMessage(
            PendingMessage(
                self,
                final_key,
                final_value,
                key_serializer=key_serializer,
                value_serializer=value_serializer,
                partition=partition,
                timestamp=timestamp,
                headers=open_headers,
                callback=callback,
                # Python 3.6.0: NamedTuple doesn't support optional fields
                # [ask]
                topic=None,
                offset=None,
            ),
        )

    def prepare_headers(
            self, headers: Optional[HeadersArg]) -> OpenHeadersArg:
        """Prepare ``headers`` passed before publishing."""
        if headers is not None:
            return prepare_headers(headers)
        return {}

    async def _send_now(
            self,
            key: K = None,
            value: V = None,
            partition: int = None,
            timestamp: float = None,
            headers: HeadersArg = None,
            schema: SchemaT = None,
            key_serializer: CodecArg = None,
            value_serializer: CodecArg = None,
            callback: MessageSentCallback = None) -> Awaitable[RecordMetadata]:
        return await self.publish_message(
            self.as_future_message(
                key, value, partition, timestamp, headers,
                schema, key_serializer, value_serializer, callback))

    async def publish_message(self, fut: FutureMessage,
                              wait: bool = True) -> Awaitable[RecordMetadata]:
        """Publish message to channel.

        This is the interface used by ``topic.send()``, etc.
        to actually publish the message on the channel
        after being buffered up or similar.

        It takes a :class:`~faust.types.FutureMessage` object,
        which contains all the information required to send
        the message, and acts as a promise that is resolved
        once the message has been fully transmitted.
        """
        event = self._future_message_to_event(fut)
        await self.put(event)
        topic, partition = tp = TP(
            fut.message.topic or '<anon>',
            fut.message.partition or -1)
        return await self._finalize_message(
            fut, RecordMetadata(
                topic=topic,
                partition=partition,
                topic_partition=tp,
                offset=-1,
                timestamp=fut.message.timestamp,
                timestamp_type=1,
            ),
        )

    def _future_message_to_event(self, fut: FutureMessage) -> EventT:
        return self._create_event(
            fut.message.key, fut.message.value, fut.message.headers,
            message=_PendingMessage_to_Message(fut.message))

    async def _finalize_message(self, fut: FutureMessage,
                                result: RecordMetadata) -> FutureMessage:
        fut.set_result(result)
        if fut.message.callback:
            await maybe_async(fut.message.callback(fut))
        return fut

    @stampede
    async def maybe_declare(self) -> None:
        """Declare/create this channel, but only if it doesn't exist."""
        ...

    async def declare(self) -> None:
        """Declare/create this channel.

        This is used to create this channel on a server,
        if that is required to operate it.
        """
        ...

    def prepare_key(
            self,
            key: K,
            key_serializer: CodecArg,
            schema: SchemaT = None,
            headers: OpenHeadersArg = None) -> Tuple[Any, OpenHeadersArg]:
        """Prepare key before it is sent to this channel.

        :class:`~faust.Topic` uses this to implement serialization of keys
        sent to the channel.
        """
        return key, headers

    def prepare_value(
            self,
            value: V,
            value_serializer: CodecArg,
            schema: SchemaT = None,
            headers: OpenHeadersArg = None) -> Tuple[Any, OpenHeadersArg]:
        """Prepare value before it is sent to this channel.

        :class:`~faust.Topic` uses this to implement serialization of values
        sent to the channel.
        """
        return value, headers

    async def decode(self, message: Message, *,
                     propagate: bool = False) -> EventT[T]:
        """Decode :class:`~faust.types.Message` into :class:`~faust.Event`."""
        return self._create_event(
            message.key, message.value, message.headers, message=message)

    async def deliver(self, message: Message) -> None:  # pragma: no cover
        """Deliver message to queue from consumer.

        This is called by the consumer to deliver the message
        to the channel.
        """
        ...  # closure compiled at __init__

    def _compile_deliver(self) -> Callable[[Message], Awaitable[None]]:
        put = None

        async def deliver(message: Message) -> None:
            nonlocal put
            if put is None:
                # NOTE circumvents self.put, using queue directly
                put = self.queue.put
            event = await self.decode(message)
            await put(event)

        return deliver

    def _create_event(self,
                      key: K,
                      value: V,
                      headers: Optional[HeadersArg],
                      message: Message) -> EventT[T]:
        return self.app.create_event(key, value, headers, message)

    async def put(self, value: EventT[T_contra]) -> None:
        """Put event onto this channel."""
        root = self._root if self._root is not None else self
        for subscriber in root._subscribers:
            await subscriber.queue.put(value)

    async def get(self, *, timeout: Seconds = None) -> EventT[T]:
        """Get the next :class:`~faust.Event` received on this channel."""
        timeout_: float = want_seconds(timeout)
        if timeout_:
            return await asyncio.wait_for(self.queue.get(), timeout=timeout_)
        return await self.queue.get()

    def empty(self) -> bool:
        """Return :const:`True` if the queue is empty."""
        return self.queue.empty()

    async def on_key_decode_error(self, exc: Exception,
                                  message: Message) -> None:
        """Unable to decode the key of an item in the queue.

        See Also:
            :meth:`on_decode_error`
        """
        await self.on_decode_error(exc, message)
        await self.throw(exc)

    async def on_value_decode_error(self, exc: Exception,
                                    message: Message) -> None:
        """Unable to decode the value of an item in the queue.

        See Also:
            :meth:`on_decode_error`
        """
        await self.on_decode_error(exc, message)
        await self.throw(exc)

    async def on_decode_error(self, exc: Exception, message: Message) -> None:
        """Signal that there was an error reading an event in the queue.

        When a message in the channel needs deserialization
        to be reconstructed back to its original form, we will sometimes
        see decoding/deserialization errors being raised, from missing
        fields or malformed payloads, and so on.

        We will log the exception, but you can also override this
        to perform additional actions.

        Admonition: Kafka
            In the event a deserialization error occurs, we
            HAVE to commit the offset of the source message to continue
            processing the stream.

            For this reason it is important that you keep a close eye on
            error logs. For easy of use, we suggest using log aggregation
            software, such as Sentry, to surface these errors to your
            operations team.
        """
        ...

    def on_stop_iteration(self) -> None:
        """Signal that iteration over this channel was stopped.

        Tip:
            Remember to call ``super`` when overriding this method.
        """
        ...

    def derive(self, **kwargs: Any) -> ChannelT[T]:
        """Derive new channel from this channel, using new configuration.

        See :class:`faust.Topic.derive`.

        For local channels this will simply return the same channel.
        """
        return self

    def __aiter__(self) -> ChannelT[T]:
        return self if self.is_iterator else self.clone(is_iterator=True)

    async def __anext__(self) -> EventT[T]:
        if not self.is_iterator:
            raise RuntimeError('Need to call channel.__aiter__()')
        return await self.queue.get()

    async def throw(self, exc: BaseException) -> None:
        """Throw exception to be received by channel subscribers.

        Tip:
            When you find yourself having to call this from
            a regular, non-``async def`` function, you can use :meth:`_throw`
            instead.
        """
        self.queue._throw(exc)

    def _throw(self, exc: BaseException) -> None:
        """Non-async version of :meth:`throw`."""
        self.queue._throw(exc)

    def __repr__(self) -> str:
        s = f'<{self.label}@{self._object_id_as_hex()}'
        if self.active_partitions is not None:
            # if we are restricted to a specific set of topic partitions,
            # then include that in repr(channel).
            if self.active_partitions:
                active = '{' + ', '.join(sorted(
                    f'{tp.topic}:{tp.partition}'
                    for tp in self.active_partitions)) + '}'
            else:
                # a defined, but empty .active_partitions signifies
                # that we are still waiting for an assignment
                # from the Consumer.
                active = '{<pending for assignment>}'
            s += f' active={active}'
        s += '>'
        return s

    def _object_id_as_hex(self) -> str:
        # hexadecimal version of id(self)
        return f'{id(self):#x}'

    def __str__(self) -> str:
        # subclasses should override this
        return '<ANON>'

    @property
    def subscriber_count(self) -> int:
        """Return number of active subscribers to local channel."""
        return len(self._subscribers)

    @property
    def label(self) -> str:
        """Short textual description of channel."""
        sym = '(*)' if self.is_iterator else ''
        return f'{sym}{type(self).__name__}: {self}'


class SerializedChannel(Channel[T]):

    def __init__(self,
                 app: AppT,
                 *,
                 schema: SchemaT = None,
                 key_type: ModelArg = None,
                 value_type: ModelArg = None,
                 key_serializer: CodecArg = None,
                 value_serializer: CodecArg = None,
                 allow_empty: bool = None,
                 **kwargs: Any) -> None:
        self.app = app  # need to set early
        if schema is not None:
            self._contribute_to_schema(
                schema,
                key_type=key_type,
                value_type=value_type,
                key_serializer=key_serializer,
                value_serializer=value_serializer,
                allow_empty=allow_empty,
            )
        else:
            schema = self._get_default_schema(
                key_type, value_type,
                key_serializer, value_serializer,
                allow_empty)

        super().__init__(
            app,
            schema=schema,
            key_type=key_type,
            value_type=value_type,
            **kwargs,
        )
        self.key_serializer = self.schema.key_serializer
        self.value_serializer = self.schema.value_serializer
        self.allow_empty = self.schema.allow_empty

    def _contribute_to_schema(self, schema: SchemaT, *,
                              key_type: ModelArg = None,
                              value_type: ModelArg = None,
                              key_serializer: CodecArg = None,
                              value_serializer: CodecArg = None,
                              allow_empty: bool = None) -> None:
        # Update schema and take compat attributes
        # from passed schema.
        schema.update(
            key_type=key_type,
            value_type=value_type,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            allow_empty=allow_empty,
        )

    def _get_default_schema(self,
                            key_type: ModelArg = None,
                            value_type: ModelArg = None,
                            key_serializer: CodecArg = None,
                            value_serializer: CodecArg = None,
                            allow_empty: bool = None) -> SchemaT:
        return self.app.conf.Schema(
            key_type=key_type,
            value_type=value_type,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            allow_empty=allow_empty,
        )

    @no_type_check  # incompatible with base class, but OK
    def _clone_args(self) -> Mapping:
        return {
            **super()._clone_args(),
            **{
                'key_serializer': self.key_serializer,
                'value_serializer': self.value_serializer,
            },
        }

    def prepare_key(
            self,
            key: K,
            key_serializer: CodecArg,
            schema: SchemaT = None,
            headers: OpenHeadersArg = None) -> Tuple[Any, OpenHeadersArg]:
        """Serialize key to format suitable for transport."""
        if key is not None:
            schema = schema or self.schema
            assert schema is not None
            return schema.dumps_key(self.app, key,
                                    serializer=key_serializer,
                                    headers=headers)
        return None, headers

    def prepare_value(
            self,
            value: V,
            value_serializer: CodecArg,
            schema: SchemaT = None,
            headers: OpenHeadersArg = None) -> Tuple[Any, OpenHeadersArg]:
        """Serialize value to format suitable for transport."""
        schema = schema or self.schema
        assert schema is not None
        return schema.dumps_value(self.app, value,
                                  serializer=value_serializer,
                                  headers=headers)



import logging
import operator
import socket
from contextlib import contextmanager
from typing import Any, Callable, Mapping, MutableMapping, Optional, cast
from mode import Service
from faust.types import AppT
from . import states
from .app import send_update
from .models import Status

CHECK_FREQUENCY = 5.0


class Check(Service):
    description: str = ''

    # This can be used to format the "current value" in error logs.
    # If not set it will just use ``repr(value)``.
    prev_value_repr: Optional[str] = None
    current_value_repr: Optional[str] = None

    state_to_severity = {
        states.SLOW: logging.WARNING,
        states.STALL: logging.ERROR,
        states.FAIL: logging.CRITICAL,
    }

    faults_to_state = [
        (6, states.STALL),
        (3, states.SLOW),
        (0, states.OK),
    ]

    default_operator = None

    def __init__(self,
                 name: str,
                 get_value: Callable[[], Any] = None,
                 operator: Callable[[Any, Any], bool] = None,
                 **kwargs: Any):
        self.name = name
        self._get_value = cast(Callable[[], Any], get_value)
        if operator is None:
            operator = self.default_operator
        self.operator = operator
        self.faults = 0
        self.prev_value = None
        self.status = states.OK
        self.interval_skew = 0.0
        self.app = None
        super().__init__(**kwargs)

    def to_representation(self, app, severity) -> Status:
        return Status(
            app_id=app.conf.id,
            hostname=socket.gethostname(),
            category=self.name,
            color=self.color,
            count=self.faults,
            state=self.status,
            severity=logging.getLevelName(severity),
        )

    def asdict(self) -> Mapping[str, Any]:
        return {
            'state': self.status,
            'color': self.color,
            'faults': self.faults,
        }

    def get_value(self) -> Any:
        if self._get_value is not None:
            return self._get_value()
        raise NotImplementedError()

    async def on_rebalancing(self, app: AppT):
        self.status = states.REBALANCING
        await send_update(app, self.to_representation(app, logging.INFO))

    async def on_unassigned(self, app: AppT):
        self.status = states.UNASSIGNED
        await send_update(app, self.to_representation(app, logging.INFO))

    async def on_paused(self, app: AppT):
        self.status = states.PAUSED
        await send_update(app, self.to_representation(app, logging.INFO))

    async def check(self, app: AppT) -> None:
        current_value: Any = self.get_value()
        prev_value = self.prev_value
        severity = app.log.info
        try:
            if prev_value is not None:
                if self.compare(prev_value, current_value):
                    self.faults += 1
                    self.status = self.get_state_for_faults(self.faults)
                    severity = self.state_to_severity.get(
                        self.status, logging.INFO)
                    await self.on_failed_log(
                        severity, app, prev_value, current_value)
                else:
                    self.faults = 0
                    self.status = states.OK
                    await self.on_ok_log(app, prev_value, current_value)
            self.store_previous_value(current_value)
        except Exception as exc:
            print(f'ERROR: {exc!r}')
            raise

    def compare(self, prev_value: Any, current_value: Any):
        return self.operator(current_value, prev_value)

    def store_previous_value(self, current_value):
        self.prev_value = current_value

    async def on_failed_log(self,
                            severity: int,
                            app: AppT,
                            prev_value: Any,
                            current_value: Any) -> None:
        await send_update(app, self.to_representation(app, severity))
        prev_value_repr = self.prev_value_repr
        current_value_repr = self.current_value_repr
        if current_value_repr is None:
            current_value_repr = repr(current_value)
        if prev_value_repr is None:
            prev_value_repr = repr(prev_value)

        app.log.log(severity,
                    '%s:%s %s (x%s): was %s now %s',
                    app.conf.id, self.name, self.negate_description,
                    self.faults, prev_value_repr, current_value_repr,
                    extra={'no_alert': True})

    async def on_ok_log(self,
                        app: AppT,
                        prev_value: Any,
                        current_value: Any) -> None:
        await send_update(app, self.to_representation(app, logging.INFO))
        app.log.info('%s:%s %s: was %s now %s',
                     app.conf.id, self.name, self.description,
                     prev_value, current_value,
                     extra={'no_alert': True})

    def get_state_for_faults(self, faults: int) -> str:
        for level, state in self.faults_to_state:
            if faults > level:
                return state
        return states.OK

    @property
    def color(self) -> str:
        if self.status in states.OK_STATES:
            return 'green'
        elif self.status in states.MAYBE_STATES:
            return 'yellow'
        return 'red'

    @Service.task
    async def _run_check(self) -> None:
        try:
            app = self.app
            while not self.should_stop:
                await self.sleep(CHECK_FREQUENCY)
                if app.system_checks.paused:
                    await self.on_paused(app)
                elif app.rebalancing:
                    await self.on_rebalancing(app)
                elif app.unassigned:
                    await self.on_unassigned(app)
                else:
                    await self.check(app)
        except Exception as exc:
            print(f'RUN CHECK RAISED: {exc!r}')
            raise

    @property
    def label(self) -> str:
        return f'{type(self).__name__}: {self.name}'


class Increasing(Check):
    default_operator = operator.le
    description = 'increasing'
    negate_description = 'not increasing'


def _transitioned_to_false(previous: bool, current: bool) -> bool:
    return not current


class Condition(Check):
    description = 'functional'
    negate_description = 'nonfunctional'

    default_operator = _transitioned_to_false

    faults_to_state = [
        (1, states.FAIL),
        (0, states.OK),
    ]


class Stationary(Check):
    """Monitors a value that should stand still, i.e, not going up or down."""
    description = 'functional'
    negate_description = 'increasing'

    default_operator = operator.ne


class SystemChecks(Service):
    checks: MutableMapping[str, Check]
    current_skew = 0.0
    paused: bool = False

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        self.app = app
        self.checks = {}
        Service.__init__(self, **kwargs)

    def on_init_dependencies(self):
        return self.checks.values()

    @contextmanager
    def pause(self):
        self.paused = True
        try:
            yield
        finally:
            self.paused = False

    def add(self, check: Check) -> None:
        self.checks[check.name] = check
        self.current_skew += 0.2
        check.interval_skew = self.current_skew
        check.app = self.app

    def remove(self, name: str) -> None:
        self.checks.pop(name, None)



from typing import Any, AsyncIterator
import faust

app = faust.App('faust.stress')
rpc_topic = app.topic('faust.stress.rpc')


@app.agent(rpc_topic, concurrency=10)
async def simple(it: AsyncIterator[Any]) -> AsyncIterator[Any]:
    async for value in it:
        yield value



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Account(models.Model):
    name = models.CharField(_('name'), max_length=100)
    score = models.DecimalField(_('score'), default=0.0,
                                max_digits=1000, decimal_places=1000)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('account')
        verbose_name_plural = _('accounts')



#!/usr/bin/env python

# This example needs pytz so that we can set the approprate timezone when
# setting the crontab

from random import random
from datetime import datetime
import faust
import pytz

app = faust.App('tz_aware', broker='kafka://localhost:9092')
pacific = pytz.timezone('US/Pacific')


class Model(faust.Record, serializer='json'):
    random: float


tz_aware_topic = app.topic('tz_aware_topic', value_type=Model)


@app.agent(tz_aware_topic)
async def consume(stream):
    async for record in stream:
        print(f'record: {record}')


@app.crontab('16 20 * * *', tz=pacific, on_leader=True)
async def publish_at_8_20_pm_pacific():
    print('-- This should be sent at 20:16 Pacific time --')
    print(f'Sending message at: {datetime.now()}')
    msg = Model(random=round(random(), 2))
    await tz_aware_topic.send(value=msg)


if __name__ == '__main__':
    app.main()



import click.exceptions
import pytest
from yarl import URL
from faust.cli.params import CaseInsensitiveChoice, TCPPort, URLParam


def test_CaseInsensitiveChoice():
    choices = CaseInsensitiveChoice([
        'FOO',
        'BAR',
        'baz',
    ])

    assert choices.convert('FOO', None, None) == 'FOO'
    assert choices.convert('foo', None, None) == 'foo'
    assert choices.convert('Foo', None, None) == 'Foo'
    assert choices.convert('BAZ', None, None) == 'BAZ'

    with pytest.raises(click.exceptions.BadParameter):
        choices.convert('xuz', None, None)


def test_TCPPort():
    port = TCPPort()
    assert port.convert(1, None, None) == 1
    assert port.convert(30, None, None) == 30
    assert port.convert(65535, None, None) == 65535
    with pytest.raises(click.exceptions.BadParameter):
        port.convert(0, None, None)
    with pytest.raises(click.exceptions.BadParameter):
        port.convert(-312, None, None)
    with pytest.raises(click.exceptions.BadParameter):
        port.convert(513412321, None, None)
    with pytest.raises(click.exceptions.BadParameter):
        port.convert(65536, None, None)


def test_URLParam():
    urlp = URLParam()
    assert repr(urlp) == 'URL'

    url = urlp.convert('http://foo.com/path/', None, None)
    assert isinstance(url, URL)
    assert url.host == 'foo.com'



import abc
import asyncio
import typing
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
)

from mode import ServiceT, SupervisorStrategyT
from mode.utils.collections import ManagedUserDict

from .codecs import CodecArg
from .core import HeadersArg, K, V
from .events import EventT
from .models import ModelArg
from .serializers import SchemaT
from .streams import StreamT
from .topics import ChannelT
from .tuples import Message, RecordMetadata, TP

if typing.TYPE_CHECKING:
    from .app import AppT as _AppT
else:
    class _AppT: ...          # noqa

__all__ = [
    'AgentErrorHandler',
    'AgentFun',
    'ActorT',
    'ActorRefT',
    'AgentManagerT',
    'AgentT',
    'AgentTestWrapperT',
    'AsyncIterableActorT',
    'AwaitableActorT',
    'ReplyToArg',
    'SinkT',
]

_T = TypeVar('_T')
AgentErrorHandler = Callable[['AgentT', BaseException], Awaitable]
AgentFun = Callable[
    [StreamT[_T]],
    Union[Coroutine[Any, Any, None], Awaitable[None], AsyncIterable],
]


#: A sink can be: Agent, Channel
#: or callable/async callable taking value as argument.
SinkT = Union['AgentT', ChannelT, Callable[[Any], Union[Awaitable, None]]]

ReplyToArg = Union['AgentT', ChannelT, str]


class ActorT(ServiceT, Generic[_T]):

    agent: 'AgentT'
    stream: StreamT
    it: _T
    actor_task: Optional[asyncio.Task]
    active_partitions: Optional[Set[TP]]

    #: If multiple instance are started for concurrency, this is its index.
    index: Optional[int] = None

    @abc.abstractmethod
    def __init__(self, agent: 'AgentT', stream: StreamT, it: _T,
                 active_partitions: Set[TP] = None,
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def cancel(self) -> None:
        ...

    @abc.abstractmethod
    async def on_isolated_partition_revoked(self, tp: TP) -> None:
        ...

    @abc.abstractmethod
    async def on_isolated_partition_assigned(self, tp: TP) -> None:
        ...

    @abc.abstractmethod
    def traceback(self) -> str:
        ...


class AsyncIterableActorT(ActorT[AsyncIterable], AsyncIterable):
    """Used for agent function that yields."""


class AwaitableActorT(ActorT[Awaitable], Awaitable):
    """Used for agent function that do not yield."""


ActorRefT = ActorT[Union[AsyncIterable, Awaitable]]


class AgentT(ServiceT, Generic[_T]):

    name: str
    app: _AppT
    concurrency: int
    help: str
    supervisor_strategy: Optional[Type[SupervisorStrategyT]]
    isolated_partitions: bool

    @abc.abstractmethod
    def __init__(self,
                 fun: AgentFun,
                 *,
                 name: str = None,
                 app: _AppT = None,
                 channel: Union[str, ChannelT] = None,
                 concurrency: int = 1,
                 sink: Iterable[SinkT] = None,
                 on_error: AgentErrorHandler = None,
                 supervisor_strategy: Type[SupervisorStrategyT] = None,
                 help: str = None,
                 schema: SchemaT = None,
                 key_type: ModelArg = None,
                 value_type: ModelArg = None,
                 isolated_partitions: bool = False,
                 **kwargs: Any) -> None:
        self.fun: AgentFun = fun

    @abc.abstractmethod
    def actor_tracebacks(self) -> List[str]:
        ...

    @abc.abstractmethod
    def __call__(self, *,
                 index: int = None,
                 active_partitions: Set[TP] = None,
                 stream: StreamT = None,
                 channel: ChannelT = None) -> ActorRefT:
        ...

    @abc.abstractmethod
    def test_context(self,
                     channel: ChannelT = None,
                     supervisor_strategy: SupervisorStrategyT = None,
                     **kwargs: Any) -> 'AgentTestWrapperT':
        ...

    @abc.abstractmethod
    def add_sink(self, sink: SinkT) -> None:
        ...

    @abc.abstractmethod
    def stream(self, **kwargs: Any) -> StreamT:
        ...

    @abc.abstractmethod
    async def on_partitions_assigned(self, assigned: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def on_partitions_revoked(self, revoked: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    async def cast(self,
                   value: V = None,
                   *,
                   key: K = None,
                   partition: int = None,
                   timestamp: float = None,
                   headers: HeadersArg = None) -> None:
        ...

    @abc.abstractmethod
    async def ask(self,
                  value: V = None,
                  *,
                  key: K = None,
                  partition: int = None,
                  timestamp: float = None,
                  headers: HeadersArg = None,
                  reply_to: ReplyToArg = None,
                  correlation_id: str = None) -> Any:
        ...

    @abc.abstractmethod
    async def send(self,
                   *,
                   key: K = None,
                   value: V = None,
                   partition: int = None,
                   timestamp: float = None,
                   headers: HeadersArg = None,
                   key_serializer: CodecArg = None,
                   value_serializer: CodecArg = None,
                   reply_to: ReplyToArg = None,
                   correlation_id: str = None) -> Awaitable[RecordMetadata]:
        ...

    @abc.abstractmethod
    @no_type_check  # XXX mypy bugs out on this
    async def map(self,
                  values: Union[AsyncIterable, Iterable],
                  key: K = None,
                  reply_to: ReplyToArg = None) -> AsyncIterator:
        ...

    @abc.abstractmethod
    @no_type_check  # XXX mypy bugs out on this
    async def kvmap(
            self,
            items: Union[AsyncIterable[Tuple[K, V]], Iterable[Tuple[K, V]]],
            reply_to: ReplyToArg = None) -> AsyncIterator[str]:
        ...

    @abc.abstractmethod
    async def join(self,
                   values: Union[AsyncIterable[V], Iterable[V]],
                   key: K = None,
                   reply_to: ReplyToArg = None) -> List[Any]:
        ...

    @abc.abstractmethod
    async def kvjoin(
            self,
            items: Union[AsyncIterable[Tuple[K, V]], Iterable[Tuple[K, V]]],
            reply_to: ReplyToArg = None) -> List[Any]:
        ...

    @abc.abstractmethod
    def info(self) -> Mapping:
        ...

    @abc.abstractmethod
    def clone(self, *, cls: Type['AgentT'] = None, **kwargs: Any) -> 'AgentT':
        ...

    @abc.abstractmethod
    def get_topic_names(self) -> Iterable[str]:
        ...

    @property
    @abc.abstractmethod
    def channel(self) -> ChannelT:
        ...

    @channel.setter
    def channel(self, channel: ChannelT) -> None:
        ...

    @property
    @abc.abstractmethod
    def channel_iterator(self) -> AsyncIterator:
        ...

    @channel_iterator.setter
    def channel_iterator(self, channel: AsyncIterator) -> None:
        ...

    @abc.abstractmethod
    def _agent_label(self, name_suffix: str = '') -> str:
        ...


class AgentManagerT(ServiceT, ManagedUserDict[str, AgentT]):
    app: _AppT

    @abc.abstractmethod
    async def wait_until_agents_started(self) -> None:
        ...

    @abc.abstractmethod
    async def on_rebalance(self,
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        ...

    @abc.abstractmethod
    def tracebacks(self) -> Mapping[str, List[str]]:
        ...

    @abc.abstractmethod
    def human_tracebacks(self) -> str:
        ...


class AgentTestWrapperT(AgentT, AsyncIterable):

    new_value_processed: asyncio.Condition
    original_channel: ChannelT
    results: MutableMapping[int, Any]
    sent_offset: int = 0
    processed_offset: int = 0

    @abc.abstractmethod
    def __init__(self,
                 *args: Any,
                 original_channel: ChannelT = None,
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    async def put(self,
                  value: V = None,
                  key: K = None,
                  partition: int = None,
                  timestamp: float = None,
                  headers: HeadersArg = None,
                  key_serializer: CodecArg = None,
                  value_serializer: CodecArg = None,
                  *,
                  reply_to: ReplyToArg = None,
                  correlation_id: str = None,
                  wait: bool = True) -> EventT:
        ...

    @abc.abstractmethod
    def to_message(self,
                   key: K,
                   value: V,
                   *,
                   partition: int = 0,
                   offset: int = 0,
                   timestamp: float = None,
                   timestamp_type: int = 0,
                   headers: HeadersArg = None) -> Message:
        ...

    @abc.abstractmethod
    async def throw(self, exc: BaseException) -> None:
        ...



import io
import json
import os
import sys
import types
from pathlib import Path
import click
import pytest
from faust.cli import AppCommand, Command, call_command
from faust.cli.base import (
    DEFAULT_LOGLEVEL,
    _Group,
    _prepare_cli,
    argument,
    compat_option,
    find_app,
    option,
)
from faust.types._env import BLOCKING_TIMEOUT, CONSOLE_PORT
from mode import Worker
from mode.utils.mocks import AsyncMock, Mock, call, patch


class test_argument:

    def test_repr(self):
        assert repr(argument(default=1))


class test_option:

    def test_repr(self):
        assert repr(option('--foo', '--bar', default=1))


def test_call_command():
    with patch('faust.cli.base.cli') as cli:
        cli.side_effect = SystemExit(3)
        exitcode, stdout, stderr = call_command('foo', ['x', 'y'])
        cli.assert_called_once_with(
            args=['foo', 'x', 'y'],
            side_effects=False,
            stdout=stdout,
            stderr=stderr,
        )
        assert exitcode == 3


def test_call_command__no_exit():
    with patch('faust.cli.base.cli') as cli:
        exitcode, stdout, stderr = call_command('foo', ['x', 'y'])
        cli.assert_called_once_with(
            args=['foo', 'x', 'y'],
            side_effects=False,
            stdout=stdout,
            stderr=stderr,
        )
        assert exitcode == 0


def test_call_command__custom_ins():
    o_out = io.StringIO()
    o_err = io.StringIO()
    with patch('faust.cli.base.cli') as cli:
        exitcode, stdout, stderr = call_command(
            'foo', ['x', 'y'],
            stdout=o_out,
            stderr=o_err,
        )
        cli.assert_called_once_with(
            args=['foo', 'x', 'y'],
            side_effects=False,
            stdout=stdout,
            stderr=stderr,
        )
        assert exitcode == 0
        assert stdout is o_out
        assert stderr is o_err


def test_compat_option():
    option = compat_option('--foo', default=1, state_key='foo')
    ctx = Mock(name='ctx')
    param = Mock(name='param')
    state = ctx.ensure_object.return_value
    state.foo = 33
    print(dir(option(ctx)))
    option(ctx)._callback(ctx, param, None) == 33
    option(ctx)._callback(ctx, param, 44) == 44
    state.foo = None
    option(ctx)._callback(ctx, param, 44) == 44


def test_find_app():
    imp = Mock(name='imp')
    symbol_by_name = Mock(name='symbol_by_name')
    with patch('faust.cli.base.prepare_app') as prepare_app:
        res = find_app('foo', imp=imp, symbol_by_name=symbol_by_name)
        assert res is prepare_app.return_value
        symbol_by_name.assert_called_once_with('foo', imp=imp)
        prepare_app.assert_called_once_with(
            symbol_by_name.return_value, 'foo')


def test_find_app__attribute_error():
    imp = Mock(name='imp')
    symbol_by_name = Mock(name='symbol_by_name')
    symbol_by_name.side_effect = AttributeError()
    with patch('faust.cli.base.prepare_app') as prepare_app:
        res = find_app('foo', imp=imp, symbol_by_name=symbol_by_name)
        assert res is prepare_app.return_value
        symbol_by_name.assert_called_once_with('foo', imp=imp)
        imp.assert_called_once_with('foo')
        prepare_app.assert_called_once_with(
            imp.return_value, 'foo')


def test_find_app__app_is_module():
    imp = Mock(name='imp')
    symbol_by_name = Mock(name='symbol_by_name')
    symbol_by_name.side_effect = AttributeError()
    imp.return_value = types.ModuleType('foo')
    imp.return_value.app = types.ModuleType('foo.app')
    with pytest.raises(AttributeError):
        find_app('foo', imp=imp, symbol_by_name=symbol_by_name)


def test_find_app__app_is_module_but_has_app():
    imp = Mock(name='imp')
    symbol_by_name = Mock(name='symbol_by_name')
    symbol_by_name.side_effect = AttributeError()
    imp.return_value = types.ModuleType('foo')
    imp.return_value.app = Mock(name='app')
    with patch('faust.cli.base.prepare_app') as prepare_app:
        find_app('foo', imp=imp, symbol_by_name=symbol_by_name)
        prepare_app.assert_called_once_with(imp.return_value.app, 'foo')


class test_Group:

    @pytest.fixture()
    def group(self):
        return _Group()

    def test_get_help(self, *, group):
        ctx = group.make_context('prog', ['x'])
        group._maybe_import_app = Mock()
        group.get_help(ctx)
        group._maybe_import_app.assert_called_once_with()

    def test_get_usage(self, *, group):
        ctx = group.make_context('prog', ['x'])
        group._maybe_import_app = Mock()
        group.get_usage(ctx)
        group._maybe_import_app.assert_called_once_with()

    @pytest.mark.parametrize('argv,expected_chdir,expected_app', [
        (['--foo', '-W', '/foo'], '/foo', None),
        (['--foo', '--workdir', '/foo'], '/foo', None),
        (['--foo', '--workdir=/foo'], '/foo', None),
        (['--foo', '-A', 'foo'], None, 'foo'),
        (['--foo', '--app', 'foo'], None, 'foo'),
        (['--foo', '--app=foo'], None, 'foo'),
        (['--foo', '--app=foo', '--workdir', '/foo'], '/foo', 'foo'),
        ([], None, None),
    ])
    def test_maybe_import_app(self, argv, expected_chdir, expected_app, *,
                              group):
        with patch('os.chdir') as chdir:
            with patch('faust.cli.base.find_app') as find_app:
                group._maybe_import_app(argv)
                if expected_chdir:
                    chdir.assert_called_once_with(
                        Path(expected_chdir).absolute())
                else:
                    chdir.assert_not_called()
                if expected_app:
                    find_app.assert_called_once_with(expected_app)
                else:
                    find_app.assert_not_called()

    def test_maybe_import_app__missing_argument(self, *, group):
        with pytest.raises(click.UsageError):
            group._maybe_import_app(['--foo', '-A'])
        with pytest.raises(click.UsageError):
            group._maybe_import_app(['--foo', '--app'])


def test__prepare_cli():
    ctx = Mock(name='context')
    state = ctx.ensure_object.return_value = Mock(name='state')
    root = ctx.find_root.return_value = Mock(name='root')
    root.side_effects = False
    app = Mock(name='app')
    try:
        _prepare_cli(
            ctx,
            app=app,
            quiet=False,
            debug=True,
            workdir='/foo',
            datadir='/data',
            json=True,
            no_color=False,
            loop='foo',
        )

        assert state.app is app
        assert not state.quiet
        assert state.debug
        assert state.workdir == '/foo'
        assert state.datadir == '/data'
        assert state.json
        assert not state.no_color
        assert state.loop == 'foo'

        root.side_effects = True
        with patch('os.chdir') as chdir:
            with patch('faust.cli.base.enable_all_colors') as eac:
                with patch('faust.utils.terminal.isatty') as isatty:
                    with patch('faust.cli.base.disable_all_colors') as dac:
                        isatty.return_value = True
                        _prepare_cli(
                            ctx,
                            app=app,
                            quiet=False,
                            debug=True,
                            workdir='/foo',
                            datadir='/data',
                            json=False,
                            no_color=False,
                            loop='foo',
                        )
                        chdir.assert_called_with(Path('/foo').absolute())
                        eac.assert_called_once_with()
                        dac.assert_not_called()
                        _prepare_cli(
                            ctx,
                            app=app,
                            quiet=False,
                            debug=True,
                            workdir='/foo',
                            datadir='/data',
                            json=True,
                            no_color=False,
                            loop='foo',
                        )
                        dac.assert_called_once_with()
                        dac.reset_mock()
                        eac.reset_mock()
                        _prepare_cli(
                            ctx,
                            app=app,
                            quiet=False,
                            debug=True,
                            workdir='/foo',
                            datadir='/data',
                            json=False,
                            no_color=True,
                            loop='foo',
                        )
                        eac.assert_not_called()
                        dac.assert_called_once_with()
                        _prepare_cli(
                            ctx,
                            app=app,
                            quiet=False,
                            debug=True,
                            workdir=None,
                            datadir=None,
                            json=False,
                            no_color=True,
                            loop='foo',
                        )
    finally:
        os.environ.pop('F_DATADIR', None)
        os.environ.pop('F_WORKDIR', None)


class test_Command:

    class TestCommand(Command):
        options = [click.option('--quiet/--no-quiet')]

    @pytest.fixture()
    def ctx(self):
        return Mock(name='ctx')

    @pytest.fixture()
    def command(self, *, ctx):
        return self.TestCommand(ctx=ctx)

    @pytest.mark.asyncio
    async def test_run(self, *, command):
        await command.run()

    @pytest.mark.asyncio
    async def test_execute(self, *, command):
        command.run = AsyncMock()
        command.on_stop = AsyncMock()
        await command.execute(2, kw=3)
        command.run.assert_called_once_with(2, kw=3)
        command.on_stop.assert_called_once_with()

    def test_parse(self, *, command):
        with pytest.raises(click.UsageError):
            assert command.parse(['foo', '--quiet'])['quiet']

    def test__parse(self, *, command):
        assert command._parse

    @pytest.mark.asyncio
    async def test_on_stop(self, *, command):
        await command.on_stop()

    def test__call__(self, *, command):
        command.run_using_worker = Mock()
        command(2, kw=2)
        command.run_using_worker.assert_called_once_with(2, kw=2)

    def test_run_using_worker(self, *, command):
        command.args = (2,)
        command.kwargs = {'kw': 1, 'yw': 6}
        command.as_service = Mock()
        command.worker_for_service = Mock()
        worker = command.worker_for_service.return_value
        worker.execute_from_commandline.side_effect = KeyError()

        with patch('asyncio.get_event_loop') as get_event_loop:
            with pytest.raises(KeyError):
                command.run_using_worker(1, kw=2)
                command.as_service.assert_called_once_with(
                    get_event_loop(), 2, 1, kw=2, yw=6)
                command.worker_for_service.assert_called_once_with(
                    command.worker_for_service.return_value,
                    get_event_loop.return_value,
                )
                worker.execute_from_commandline.assert_called_once_with()

    def test_on_worker_created(self, *, command):
        assert command.on_worker_created(Mock(name='worker')) is None

    def test_as_service(self, *, command):
        loop = Mock()
        command.execute = Mock()
        with patch('faust.cli.base.Service') as Service:
            res = command.as_service(loop, 1, kw=2)
            assert res is Service.from_awaitable.return_value
            Service.from_awaitable.assert_called_once_with(
                command.execute.return_value,
                name=type(command).__name__,
                loop=loop,
            )
            command.execute.assert_called_once_with(1, kw=2)

    def test_worker_for_service(self, *, command):
        with patch('faust.cli.base.Worker') as Worker:
            service = Mock(name='service')
            loop = Mock(name='loop')
            res = command.worker_for_service(service, loop)
            assert res is Worker.return_value
            Worker.assert_called_once_with(
                service,
                debug=command.debug,
                quiet=command.quiet,
                stdout=command.stdout,
                stderr=command.stderr,
                loglevel=command.loglevel,
                logfile=command.logfile,
                blocking_timeout=command.blocking_timeout,
                console_port=command.console_port,
                redirect_stdouts=command.redirect_stdouts or False,
                redirect_stdouts_level=command.redirect_stdouts_level,
                loop=loop,
                daemon=command.daemon,
            )

    def test__Worker(self, *, command):
        assert command._Worker is Worker

    def test_tabulate__when_text(self, *, command):
        command.json = False
        data = [
            ['A', 'B', 'C'],
            ['D', 'E', 'F'],
        ]
        headers = ['a', 'b', 'c']
        assert command.tabulate(data, headers=headers)
        assert command.tabulate(data, headers=None)
        assert command.tabulate(data, headers=None, wrap_last_row=False)

    def test_tabulate__when_json(self, *, command):
        command.json = True
        data = [
            ['A', 'B', 'C'],
            ['D', 'E', 'F'],
        ]
        headers = ['a', 'b', 'c']
        assert json.loads(command.tabulate(data, headers=headers)) == [
            {'a': 'A',
             'b': 'B',
             'c': 'C'},
            {'a': 'D',
             'b': 'E',
             'c': 'F'},
        ]

    def test_tabulate_json(self, *, command):
        data = [
            ['A', 'B', 'C'],
            ['D', 'E', 'F'],
        ]
        assert json.loads(command._tabulate_json(data, headers=None)) == data

    def test_tabulate_json__headers(self, *, command):
        data = [
            ['A', 'B', 'C'],
            ['D', 'E', 'F'],
        ]
        headers = ['a', 'b', 'c']
        assert json.loads(command._tabulate_json(data, headers=headers)) == [
            {'a': 'A',
             'b': 'B',
             'c': 'C'},
            {'a': 'D',
             'b': 'E',
             'c': 'F'},
        ]

    def test_table(self, *, command):
        with patch('faust.utils.terminal.table') as table:
            data = [['A', 'B', 'C']]
            t = command.table(data, title='foo')
            assert t is table.return_value
            table.assert_called_once_with(
                data, title='foo', target=sys.stdout)

    def test_color(self, *, command):
        assert command.color('blue', 'text')

    def test_dark(self, *, command):
        assert command.dark('text')

    def test_bold(self, *, command):
        assert command.bold('text')

    def test_bold_tail(self, *, command):
        assert command.bold_tail('foo.bar.baz')

    def test_table_wrap(self, *, command):
        table = command.table([['A', 'B', 'C']])
        assert command._table_wrap(table, 'fooawqe' * 100)

    def test_say(self, *, command):
        with patch('faust.cli.base.echo') as echo:
            command.quiet = True
            command.say('foo')
            echo.assert_not_called()

            command.quiet = False
            command.say('foo')
            echo.assert_called_once_with(
                'foo', file=command.stdout, err=command.stderr)

    def test_carp(self, *, command):
        command.say = Mock()
        command.debug = False
        command.carp('foo')
        command.say.assert_not_called()
        command.debug = True
        command.carp('foo')
        command.say.assert_called_once()

    def test_dumps(self, *, command):
        with patch('faust.utils.json.dumps') as dumps:
            obj = Mock(name='obj')
            assert command.dumps(obj) is dumps.return_value
            dumps.assert_called_once_with(obj)

    def test_loglevel(self, *, command, ctx):
        assert command.loglevel == ctx.ensure_object().loglevel
        command.loglevel = 'FOO'
        assert command.loglevel == 'FOO'
        command.loglevel = None
        assert command.loglevel == DEFAULT_LOGLEVEL

    def test_blocking_timeout(self, *, command, ctx):
        assert command.blocking_timeout == ctx.ensure_object().blocking_timeout
        command.blocking_timeout = 32.41
        assert command.blocking_timeout == 32.41
        command.blocking_timeout = None
        assert command.blocking_timeout == BLOCKING_TIMEOUT

    def test_console_port(self, *, command, ctx):
        assert command.console_port == ctx.ensure_object().console_port
        command.console_port = 3241
        assert command.console_port == 3241
        command.console_port = None
        assert command.console_port == CONSOLE_PORT


class test_AppCommand:

    @pytest.fixture()
    def ctx(self):
        return Mock(name='ctx')

    @pytest.fixture()
    def command(self, *, app, ctx):
        return AppCommand(app=app, ctx=ctx)

    def test_finalize_app__str(self, *, command):
        command._app_from_str = Mock()
        command.state.app = 'foo'
        res = command._finalize_app(None)
        assert res is command._app_from_str.return_value
        command._app_from_str.assert_called_once_with('foo')

    def test_finalize_app__concrete(self, *, command, app):
        command._finalize_concrete_app = Mock()
        res = command._finalize_app(app)
        assert res is command._finalize_concrete_app.return_value
        command._finalize_concrete_app.assert_called_once_with(app)

    def test_app_from_str(self, *, command):
        with patch('faust.cli.base.find_app') as find_app:
            res = command._app_from_str('foo')
            assert res is find_app.return_value
            find_app.assert_called_once_with('foo')

    def test_app_from_str__empty(self, *, command):
        command.require_app = False
        assert command._app_from_str(None) is None
        command.require_app = True
        with pytest.raises(command.UsageError):
            command._app_from_str(None)

    def test_finalize_concrete_app(self, *, command, app):
        with patch('sys.argv', []):
            command._finalize_concrete_app(app)

    @pytest.mark.asyncio
    async def test_on_stop(self, *, command, ctx):
        app = ctx.find_root().app
        app._producer = None
        app._http_client = None
        app.started = False
        await command.on_stop()

        app._producer = Mock(stop=AsyncMock())
        app._maybe_close_http_client = AsyncMock()
        app._http_client = Mock()
        app.stop = AsyncMock()
        app.started = True

        await command.on_stop()
        app._producer.stop.assert_called_once_with()
        app.stop.assert_called_once_with()
        app._maybe_close_http_client.assert_called_once_with()

    def test_to_key(self, *, command):
        command.to_model = Mock()
        res = command.to_key('typ', 'key')
        assert res is command.to_model.return_value
        command.to_model.assert_called_once_with(
            'typ', 'key', command.key_serializer)

    def test_to_value(self, *, command):
        command.to_model = Mock()
        res = command.to_value('typ', 'value')
        assert res is command.to_model.return_value
        command.to_model.assert_called_once_with(
            'typ', 'value', command.value_serializer)

    def test_to_model(self, *, command):
        command.import_relative_to_app = Mock()
        res = command.to_model('typ', 'value', 'json')
        command.import_relative_to_app.assert_called_once_with('typ')
        model = command.import_relative_to_app.return_value
        assert res is model.loads.return_value
        model.loads.assert_called_once_with(b'value', serializer='json')

    def test_to_model__bytes(self, *, command):
        assert command.to_model(None, 'value', 'json') == b'value'

    def test_import_relative_to_app(self, *, command, ctx):
        with patch('faust.cli.base.symbol_by_name') as symbol_by_name:
            res = command.import_relative_to_app('foo')
            assert res is symbol_by_name.return_value
            symbol_by_name.assert_called_once_with('foo')

    def test_import_relative_to_app__no_origin(self, *, command, ctx):
        app = ctx.find_root().app
        app.conf.origin = None
        with patch('faust.cli.base.symbol_by_name') as symbol_by_name:
            symbol_by_name.side_effect = ImportError()
            with pytest.raises(ImportError):
                command.import_relative_to_app('foo')

    def test_import_relative_to_app__with_origin(self, *, command, ctx):
        app = ctx.find_root().app
        app.conf.origin = 'root.moo:bar'
        with patch('faust.cli.base.symbol_by_name') as symbol_by_name:
            symbol_by_name.side_effect = ImportError()
            with pytest.raises(ImportError):
                command.import_relative_to_app('foo')
            symbol_by_name.assert_has_calls([
                call('foo'),
                call('root.moo.models.foo'),
                call('root.moo.foo'),
            ])

    def test_import_relative_to_app__with_origin_l1(self, *, command, ctx):
        app = ctx.find_root().app
        app.conf.origin = 'root.moo:bar'
        with patch('faust.cli.base.symbol_by_name') as symbol_by_name:
            symbol_by_name.side_effect = [ImportError(), ImportError(), 'x']
            assert command.import_relative_to_app('foo') == 'x'
            symbol_by_name.assert_has_calls([
                call('foo'),
                call('root.moo.models.foo'),
                call('root.moo.foo'),
            ])

    def test_import_relative_to_app__with_origin_l2(self, *, command, ctx):
        app = ctx.find_root().app
        app.conf.origin = 'root.moo:bar'
        with patch('faust.cli.base.symbol_by_name') as symbol_by_name:
            symbol_by_name.side_effect = [ImportError(), 'x']
            assert command.import_relative_to_app('foo') == 'x'
            symbol_by_name.assert_has_calls([
                call('foo'),
                call('root.moo.models.foo'),
            ])

    def test_to_topic__missing(self, *, command):
        with pytest.raises(command.UsageError):
            command.to_topic(None)

    def test_to_topic__agent_prefix(self, *, command):
        command.import_relative_to_app = Mock()
        res = command.to_topic('@agent')
        assert res is command.import_relative_to_app.return_value
        command.import_relative_to_app.assert_called_once_with('agent')

    def test_to_topic__topic_name(self, *, command, ctx):
        app = ctx.find_root().app
        app.topic = Mock()
        assert command.to_topic('foo') is app.topic.return_value
        app.topic.assert_called_once_with('foo')

    def test_abbreviate_fqdn(self, *, command, ctx):
        with patch('mode.utils.text.abbr_fqdn') as abbr_fqdn:
            ret = command.abbreviate_fqdn('foo.bar.baz', prefix='prefix')
            assert ret is abbr_fqdn.return_value
            abbr_fqdn.assert_called_once_with(
                ctx.find_root().app.conf.origin,
                'foo.bar.baz',
                prefix='prefix',
            )

    def test_abbreviate_fqdn__no_origin(self, *, command, ctx):
        command.app.conf.origin = None
        assert command.abbreviate_fqdn('foo') == ''

    def test_from_handler_no_params(self, *, command):
        @command.from_handler()
        async def takes_no_args():
            ...

        @command.from_handler()
        async def takes_self_arg(self):
            ...



import pytest
from faust.tables.objects import ChangeloggedObjectManager
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock, patch

TP1 = TP('foo', 3)


@pytest.fixture()
def key():
    return Mock(name='key')


@pytest.fixture()
def table():
    return Mock(
        name='table',
        _new_store_by_url=Mock(
            return_value=Mock(
                on_rebalance=AsyncMock(),
            ),
        ),
    )


@pytest.yield_fixture()
def current_event():
    with patch('faust.tables.objects.current_event') as current_event:
        yield current_event


class ValueType:

    def __init__(self, man, key):
        self.man = man
        self.key = key
        self.synced = set()
        self.changes = []

    def sync_from_storage(self, value):
        self.synced.add(value)

    def as_stored_value(self):
        return f'{self.key}-stored'

    def apply_changelog_event(self, operation, value):
        self.changes.append((operation, value))


class test_ChangeloggedObjectManager:

    @pytest.fixture()
    def man(self, *, table):
        man = ChangeloggedObjectManager(table)
        man.ValueType = ValueType
        return man

    @pytest.fixture()
    def storage(self, *, table):
        return table._new_store_by_url.return_value

    def test_send_changelog_event(self, *, man, table, key, current_event):
        man.send_changelog_event(key, 3, 'value')
        assert key in man._dirty
        table._send_changelog.assert_called_once_with(
            current_event(), (3, key), 'value',
        )

    def test__getitem__(self, *, man):
        v1 = man['k']
        v2 = man['k']
        assert v1 is v2
        v3 = man['j']
        assert v3 is not v1

        assert man.data['k'].man is man
        assert man.data['k'].key == 'k'
        assert man.data['j'].man is man
        assert man.data['j'].key == 'j'

    def test__setitem__(self, *, man):
        with pytest.raises(NotImplementedError):
            man['k'] = 3

    def test__delitem__(self, *, man):
        with pytest.raises(NotImplementedError):
            del(man['k'])

    def test_table_type_name(self, *, man):
        assert man._table_type_name

    @pytest.mark.asyncio
    async def test_on_start(self, *, man):
        man.add_runtime_dependency = AsyncMock()
        await man.on_start()
        man.add_runtime_dependency.assert_called_once_with(man.storage)

    @pytest.mark.asyncio
    async def test_on_stop(self, *, man):
        man.flush_to_storage = Mock()
        await man.on_stop()
        man.flush_to_storage.assert_called_once_with()

    def test_persisted_offset(self, *, man, storage):
        ret = man.persisted_offset(TP1)
        storage.persisted_offset.assert_called_once_with(TP1)
        assert ret is storage.persisted_offset()

    def test_set_persisted_offset(self, *, man, storage):
        man.set_persisted_offset(TP1, 3003)
        storage.set_persisted_offset.assert_called_once_with(TP1, 3003)

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, man, storage, table):
        await man.on_rebalance(table, {TP1}, {TP1}, {TP1})
        man.storage.on_rebalance.assert_called_once_with(
            table, {TP1}, {TP1}, {TP1},
        )

    @pytest.mark.asyncio
    async def test_on_recovery_completed(self, *, man):
        man.sync_from_storage = Mock()
        await man.on_recovery_completed({TP1}, {TP1})
        man.sync_from_storage.assert_called_once_with()

    def test_sync_from_storage(self, *, man, storage):
        storage.items.return_value = [('foo', 1), ('bar', 2)]
        man.sync_from_storage()
        assert 1 in man['foo'].synced
        assert 2 in man['bar'].synced

    def test_flush_to_storage(self, *, man):
        man._storage = {}
        man._dirty = {'foo', 'bar'}
        assert man['foo']
        assert man['bar']
        man.flush_to_storage()
        assert man._storage['foo'] == 'foo-stored'

    def test_reset_state(self, *, man, storage):
        man.reset_state()
        storage.reset_state.assert_called_once_with()

    def test_apply_changelog_batch__key_is_None(self, *, man):
        event1 = Mock(name='event1')
        event1.key = None
        with pytest.raises(RuntimeError):
            man.apply_changelog_batch([event1], Mock(), Mock())

    def test_apply_changelog_batch__empty(self, *, man):
        man.apply_changelog_batch([], lambda k: k, lambda v: v)

    def test_apply_changelog_batch(self, *, man):
        event1 = Mock(name='event1')
        event1.key = 3, 'k'
        event1.value = 'foo'
        man.apply_changelog_batch([event1], lambda k: k, lambda v: v)
        assert (3, 'foo') in man['k'].changes



import asyncio
import operator
from copy import copy

import pytest
from faust import joins
from faust import Event, Record, Stream, Topic
from faust.exceptions import PartitionsMismatch
from faust.stores.base import Store
from faust.tables.base import Collection
from faust.types import TP
from faust.windows import Window
from mode import label, shortlabel
from mode.utils.mocks import AsyncMock, Mock, call, patch

TP1 = TP('foo', 0)


class User(Record):
    id: str
    name: str


class MyTable(Collection):

    def __post_init__(self, *args, **kwargs):
        self.datas = {}

    def _has_key(self, key):
        return key in self.datas

    def _get_key(self, key):
        return self.datas.get(key)

    def _set_key(self, key, value):
        self.datas[key] = value

    def _del_key(self, key):
        self.datas.pop(key, None)

    def hopping(self, *args, **kwargs):
        raise NotImplementedError()

    def tumbling(self, *args, **kwargs):
        raise NotImplementedError()

    def using_window(self, *args, **kwargs):
        raise NotImplementedError()

    def as_ansitable(self, *args, **kwargs):
        raise NotImplementedError()


class test_Collection:

    @pytest.fixture
    def table(self, *, app):
        return MyTable(app, name='name')

    def test_key_type_bytes_implies_raw_serializer(self, *, app):
        table = MyTable(app, name='name', key_type=bytes)
        assert table.key_serializer == 'raw'

    @pytest.mark.asyncio
    async def test_init_on_recover(self, *, app):
        on_recover = AsyncMock(name='on_recover')
        t = MyTable(app, name='name', on_recover=on_recover)
        assert on_recover in t._recover_callbacks
        await t.call_recover_callbacks()
        on_recover.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_recovery_completed(self, *, table):
        table.call_recover_callbacks = AsyncMock()
        await table.on_recovery_completed(set(), set())
        table.call_recover_callbacks.assert_called_once_with()

    def test_hash(self, *, table):
        assert hash(table)

    @pytest.mark.asyncio
    async def test_on_start(self, *, table):
        table.changelog_topic = Mock(
            name='changelog_topic',
            autospec=Topic,
            maybe_declare=AsyncMock(),
        )
        await table.on_start()
        table.changelog_topic.maybe_declare.assert_called_once_with()

    def test_info(self, *, table):
        assert table.info() == {
            'app': table.app,
            'name': table.name,
            'store': table._store,
            'default': table.default,
            'schema': table.schema,
            'key_type': table.key_type,
            'value_type': table.value_type,
            'changelog_topic': table._changelog_topic,
            'window': table.window,
            'extra_topic_configs': table.extra_topic_configs,
            'on_changelog_event': table._on_changelog_event,
            'recover_callbacks': table._recover_callbacks,
            'partitions': table.partitions,
            'recovery_buffer_size': table.recovery_buffer_size,
            'standby_buffer_size': table.standby_buffer_size,
            'use_partitioner': table.use_partitioner,
        }

    def test_persisted_offset(self, *, table):
        data = table._data = Mock(name='_data')
        assert table.persisted_offset(TP1) == data.persisted_offset()

    @pytest.mark.asyncio
    async def test_need_active_standby_for(self, *, table):
        table._data = Mock(
            name='_data',
            autospec=Store,
            need_active_standby_for=AsyncMock(),
        )
        assert (await table.need_active_standby_for(TP1) ==
                table._data.need_active_standby_for.coro())

    def test_reset_state(self, *, table):
        data = table._data = Mock(name='_data', autospec=Store)
        table.reset_state()
        data.reset_state.assert_called_once_with()

    def test_send_changelog(self, *, table):
        table.changelog_topic.send_soon = Mock(name='send_soon')
        event = Mock(name='event')
        table._send_changelog(event, 'k', 'v')
        table.changelog_topic.send_soon.assert_called_once_with(
            key='k',
            value='v',
            partition=event.message.partition,
            key_serializer='json',
            value_serializer='json',
            callback=table._on_changelog_sent,
            eager_partitioning=True,
        )

    def test_send_changelog__custom_serializers(self, *, table):
        event = Mock(name='event')
        table.changelog_topic.send_soon = Mock(name='send_soon')
        table._send_changelog(
            event, 'k', 'v',
            key_serializer='raw',
            value_serializer='raw',
        )
        table.changelog_topic.send_soon.assert_called_once_with(
            key='k',
            value='v',
            partition=event.message.partition,
            key_serializer='raw',
            value_serializer='raw',
            callback=table._on_changelog_sent,
            eager_partitioning=True,
        )

    def test_send_changelog__no_current_event(self, *, table):
        with pytest.raises(RuntimeError):
            table._send_changelog(None, 'k', 'v')

    def test_on_changelog_sent(self, *, table):
        fut = Mock(name='future', autospec=asyncio.Future)
        table._data = Mock(name='data', autospec=Store)
        table._on_changelog_sent(fut)
        table._data.set_persisted_offset.assert_called_once_with(
            fut.result().topic_partition, fut.result().offset,
        )

    def test_on_changelog_sent__transactions(self, *, table):
        table.app.in_transaction = True
        table.app.tables = Mock(name='tables')
        fut = Mock(name='fut')
        table._on_changelog_sent(fut)
        table.app.tables.persist_offset_on_commit.assert_called_once_with(
            table.data, fut.result().topic_partition, fut.result().offset,
        )

    def test_del_old_keys__empty(self, *, table):
        table.window = Mock(name='window')
        table._del_old_keys()

    def test_del_old_keys(self, *, table):
        on_window_close = table._on_window_close = Mock(name='on_window_close')

        table.window = Mock(name='window')
        table._data = {
            ('boo', (1.1, 1.4)): 'BOO',
            ('moo', (1.4, 1.6)): 'MOO',
            ('faa', (1.9, 2.0)): 'FAA',
            ('bar', (4.1, 4.2)): 'BAR',
        }
        table._partition_timestamps = {
            TP1: [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        }
        table._partition_timestamp_keys = {
            (TP1, 2.0): [
                ('boo', (1.1, 1.4)),
                ('moo', (1.4, 1.6)),
                ('faa', (1.9, 2.0)),
            ],
            (TP1, 5.0): [
                ('bar', (4.1, 4.2)),
            ],
        }

        def get_stale(limit):
            def is_stale(timestamp, latest_timestamp):
                return timestamp < limit
            return is_stale

        table.window.stale.side_effect = get_stale(4.0)

        table._del_old_keys()

        assert table._partition_timestamps[TP1] == [
            4.0, 5.0, 6.0, 7.0]
        assert table.data == {('bar', (4.1, 4.2)): 'BAR'}

        on_window_close.assert_has_calls([
            call(('boo', (1.1, 1.4)), 'BOO'),
            call(('moo', (1.4, 1.6)), 'MOO'),
            call(('faa', (1.9, 2.0)), 'FAA'),
        ])

        table.last_closed_window = 8.0
        table.window.stale.side_effect = get_stale(6.0)

        table._del_old_keys()

        assert not table.data

    def test_on_window_close__default(self, *, table):
        assert table._on_window_close is None
        table.on_window_close(('boo', (1.1, 1.4)), 'BOO')

    @pytest.mark.parametrize('source_n,change_n,expect_error', [
        (3, 3, False),
        (3, None, False),
        (None, 3, False),
        (3, 6, True),
        (6, 3, True),
    ])
    def test__verify_source_topic_partitions(
            self, source_n, change_n, expect_error, *, app, table):
        event = Mock(name='event', autospec=Event)
        tps = {
            event.message.topic: source_n,
            table.changelog_topic.get_topic_name(): change_n,
        }
        app.consumer.topic_partitions = Mock(side_effect=tps.get)
        if expect_error:
            with pytest.raises(PartitionsMismatch):
                table._verify_source_topic_partitions(event.message.topic)
        else:
            table._verify_source_topic_partitions(event.message.topic)

    @pytest.mark.asyncio
    async def test_clean_data(self, *, table):
        table._should_expire_keys = Mock(name='_should_expire_keys')
        table._should_expire_keys.return_value = False
        await table._clean_data(table)

        table._should_expire_keys.return_value = True
        table._del_old_keys = Mock(name='_del_old_keys')

        def on_sleep(secs, **kwargs):
            if table.sleep.call_count > 2:
                table._stopped.set()
        table.sleep = AsyncMock(name='sleep', side_effect=on_sleep)

        await table._clean_data(table)

        table._del_old_keys.assert_called_once_with()
        table.sleep.assert_called_with(
            pytest.approx(table.app.conf.table_cleanup_interval, rel=1e-1))

    def test_should_expire_keys(self, *, table):
        table.window = None
        assert not table._should_expire_keys()
        table.window = Mock(name='window', autospec=Window)
        table.window.expires = 3600
        assert table._should_expire_keys()

    def test_join(self, *, table):
        table._join = Mock(name='join')
        ret = table.join(User.id, User.name)
        table._join.assert_called_once_with(
            joins.RightJoin(stream=table, fields=(User.id, User.name)),
        )
        assert ret is table._join()

    def test_left_join(self, *, table):
        table._join = Mock(name='join')
        ret = table.left_join(User.id, User.name)
        table._join.assert_called_once_with(
            joins.LeftJoin(stream=table, fields=(User.id, User.name)),
        )
        assert ret is table._join()

    def test_inner_join(self, *, table):
        table._join = Mock(name='join')
        ret = table.inner_join(User.id, User.name)
        table._join.assert_called_once_with(
            joins.InnerJoin(stream=table, fields=(User.id, User.name)),
        )
        assert ret is table._join()

    def test_outer_join(self, *, table):
        table._join = Mock(name='join')
        ret = table.outer_join(User.id, User.name)
        table._join.assert_called_once_with(
            joins.OuterJoin(stream=table, fields=(User.id, User.name)),
        )
        assert ret is table._join()

    def test__join(self, *, table):
        with pytest.raises(NotImplementedError):
            table._join(Mock(name='join_strategy', autospec=joins.Join))

    def test_clone(self, *, table):
        t2 = table.clone()
        assert t2.info() == table.info()

    def test_combine(self, *, table):
        with pytest.raises(NotImplementedError):
            table.combine(Mock(name='joinable', autospec=Stream))

    def test_contribute_to_stream(self, *, table):
        table.contribute_to_stream(Mock(name='stream', autospec=Stream))

    @pytest.mark.asyncio
    async def test_remove_from_stream(self, *, table):
        await table.remove_from_stream(Mock(name='stream', autospec=Stream))

    def test_new_changelog_topic__window_expires(self, *, table):
        table.window = Mock(name='window', autospec=Window)
        table.window.expires = 3600.3
        assert table._new_changelog_topic(retention=None).retention == 3600.3

    def test_new_changelog_topic__default_compacting(self, *, table):
        table._changelog_compacting = True
        assert table._new_changelog_topic(compacting=None).compacting
        table._changelog_compacting = False
        assert not table._new_changelog_topic(compacting=None).compacting
        assert table._new_changelog_topic(compacting=True).compacting

    def test_new_changelog_topic__default_deleting(self, *, table):
        table._changelog_deleting = True
        assert table._new_changelog_topic(deleting=None).deleting
        table._changelog_deleting = False
        assert not table._new_changelog_topic(deleting=None).deleting
        assert table._new_changelog_topic(deleting=True).deleting

    def test_copy(self, *, table):
        assert copy(table).info() == table.info()

    def test_and(self, *, table):
        with pytest.raises(NotImplementedError):
            table & table

    def test__maybe_set_key_ttl(self, *, table):
        table._should_expire_keys = Mock(return_value=False)
        table._maybe_set_key_ttl(('k', (100, 110)), 0)

        table._should_expire_keys = Mock(return_value=True)
        table._maybe_set_key_ttl(('k', (100, 110)), 0)

    def test__maybe_del_key_ttl(self, *, table):
        table._partition_timestamp_keys[(0, 110)] = None

        table._should_expire_keys = Mock(return_value=False)
        table._maybe_del_key_ttl(('k', (100, 110)), 0)

        table._should_expire_keys = Mock(return_value=True)
        table._maybe_del_key_ttl(('k', (100, 110)), 0)

        table._partition_timestamp_keys[(0, 110)] = {
            ('k', (100, 110)),
            ('v', (100, 110)),
        }
        table._maybe_del_key_ttl(('k', (100, 110)), 0)

        assert table._partition_timestamp_keys[(0, 110)] == {('v', (100, 110))}

    def test_apply_window_op(self, *, table):
        self.mock_ranges(table)
        table._set_key(('k', 1.1), 30)
        table._set_key(('k', 1.2), 40)
        table._set_key(('k', 1.3), 50)
        table._apply_window_op(operator.add, 'k', 12, 300.3)

        assert table._get_key(('k', 1.1)) == 42
        assert table._get_key(('k', 1.2)) == 52
        assert table._get_key(('k', 1.3)) == 62

    def test_set_del_windowed(self, *, table):
        ranges = self.mock_ranges(table)
        table._set_windowed('k', 11, 300.3)
        for r in ranges:
            assert table._get_key(('k', r)) == 11
        table._del_windowed('k', 300.3)
        for r in ranges:
            assert table._get_key(('k', r)) is None

    def test_window_ranges(self, *, table):
        table.window = Mock(name='window', autospec=Window)
        table.window.ranges.return_value = [1, 2, 3]
        assert list(table._window_ranges(300.3)) == [1, 2, 3]

    def mock_ranges(self, table, ranges=[1.1, 1.2, 1.3]):  # noqa
        table._window_ranges = Mock(name='_window_ranges')
        table._window_ranges.return_value = ranges
        return ranges

    def test_relative_now(self, *, table):
        event = Mock(name='event', autospec=Event)
        table._partition_latest_timestamp[event.message.partition] = 30.3
        assert table._relative_now(event) == 30.3

    def test_relative_now__no_event(self, *, table):
        with patch('faust.tables.base.current_event') as ce:
            ce.return_value = None
            with patch('time.time') as time:
                assert table._relative_now(None) is time()

    def test_relative_event(self, *, table):
        event = Mock(name='event', autospec=Event)
        assert table._relative_event(event) is event.message.timestamp

    def test_relative_event__raises_if_no_event(self, *, table):
        with patch('faust.tables.base.current_event') as current_event:
            current_event.return_value = None
            with pytest.raises(RuntimeError):
                table._relative_event(None)

    def test_relative_field(self, *, table):
        user = User('foo', 'bar')
        event = Mock(name='event', autospec=Event)
        event.value = user
        assert table._relative_field(User.id)(event) == 'foo'

    def test_relative_field__raises_if_no_event(self, *, table):
        with pytest.raises(RuntimeError):
            table._relative_field(User.id)(event=None)

    def test_relative_timestamp(self, *, table):
        assert table._relative_timestamp(303.3)(
            Mock(name='event', autospec=Event)) == 303.3

    def test_windowed_now(self, *, table):
        with patch('faust.tables.base.current_event'):
            table.window = Mock(name='window', autospec=Window)
            table.window.earliest.return_value = 42
            table._get_key = Mock(name='_get_key')
            table._windowed_now('k')
            table._get_key.assert_called_once_with(('k', 42))

    def test_windowed_timestamp(self, *, table):
        table.window = Mock(name='window', autospec=Window)
        table.window.current.return_value = 10.1
        assert not table._windowed_contains('k', 303.3)
        table._set_key(('k', 10.1), 101.1)
        assert table._windowed_timestamp('k', 303.3) == 101.1
        assert table._windowed_contains('k', 303.3)

    def test_windowed_delta(self, *, table):
        event = Mock(name='event', autospec=Event)
        table.window = Mock(name='window', autospec=Window)
        table.window.delta.return_value = 10.1
        table._set_key(('k', 10.1), 101.1)
        assert table._windowed_delta('k', 303.3, event=event) == 101.1

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, table):
        table._data = Mock(
            name='data',
            autospec=Store,
            on_rebalance=AsyncMock(),
        )
        await table.on_rebalance({TP1}, set(), set())
        table._data.on_rebalance.assert_called_once_with(
            table, {TP1}, set(), set())

    @pytest.mark.asyncio
    async def test_on_changelog_event(self, *, table):
        event = Mock(name='event', autospec=Event)
        table._on_changelog_event = None
        await table.on_changelog_event(event)
        table._on_changelog_event = AsyncMock(name='callback')
        await table.on_changelog_event(event)
        table._on_changelog_event.assert_called_once_with(event)

    def test_label(self, *, table):
        assert label(table)

    def test_shortlabel(self, *, table):
        assert shortlabel(table)

    def test_apply_changelog_batch(self, *, table):
        table._data = Mock(name='data', autospec=Store)
        table.apply_changelog_batch([1, 2, 3])
        table._data.apply_changelog_batch.assert_called_once_with(
            [1, 2, 3],
            to_key=table._to_key,
            to_value=table._to_value,
        )

    def test_to_key(self, *, table):
        assert table._to_key([1, 2, 3]) == (1, 2, 3)
        assert table._to_key(1) == 1

    def test_to_value(self, *, table):
        v = Mock(name='v')
        assert table._to_value(v) is v

    def test__human_channel(self, *, table):
        assert table._human_channel()

    def test_repr_info(self, *, table):
        assert table._repr_info() == table.name

    def test_partition_for_key__partitioner(self, *, table, app):
        table.use_partitioner = True

        partition = None
        assert table.partition_for_key('k') is partition



# -*- coding: utf-8 -*-
"""Python Stream processing."""
# :copyright: (c) 2017-2020, Robinhood Markets, Inc.
#             All rights reserved.
# :license:   BSD (3 Clause), see LICENSE for more details.

# -- Faust is a Python stream processing library
# mainly used with Kafka, but is generic enough to be used for general
# agent and channel based programming.

# If you are here to read the code, we suggest you start with:
#
#  faust/app.py            - Faust Application
#  faust/channels.py       - Channels for communication.
#  faust/topics.py         - A topic is a named channel (e.g. a Kafka topic)
#  faust/streams.py        - The stream iterates over events in a channel.
#  faust/tables.py         - Data is stored in tables.
#  faust/agents.py         - Agents use all of the above.
# --- ~~~~~ ~ ~  ~           ~             ~   ~                   ~
import os
import re
import sys
import typing

from typing import Any, Mapping, NamedTuple, Optional, Sequence, Tuple

__version__ = '1.16.8'
__author__ = 'Robinhood Markets, Inc.'
__contact__ = 'contact@fauststream.com'
__homepage__ = 'http://faust.readthedocs.io/'
__docformat__ = 'restructuredtext'

# -eof meta-


class version_info_t(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: str


# bumpversion can only search for {current_version}
# so we have to parse the version here.
_match = re.match(r'(\d+)\.(\d+).(\d+)(.+)?', __version__)
if _match is None:  # pragma: no cover
    raise RuntimeError('THIS IS A BROKEN RELEASE!')
_temp = _match.groups()
VERSION = version_info = version_info_t(
    int(_temp[0]), int(_temp[1]), int(_temp[2]), _temp[3] or '', '')
del(_match)
del(_temp)
del(re)


# This is here to support setting the --datadir argument
# when executing a faust application module directly:
#    $ python examples/word_count.py
#
# When using the ``faust`` command, the entry point is controlled by
# .cli.base.cli.  It sets the F_DATADIR environment variable before the
# app is imported, which means that when the faust.app module is imported
# the App will have the correct default paths.
#
# However, when examples/word_count.py is executed the app is instantiated
# before the CLI, so we have to use this hack to to set the
# environment variable as early as possible.
#
# A side effect is that anything importing faust
# that also happens to have a '-D' argument in sys.argv will have the
# F_DATADIR environment variable set.  I think that's ok. [ask]
def _extract_arg_from_argv(  # pragma: no cover
        argv: Sequence[str] = sys.argv,
        *,
        shortopts: Tuple[str, ...] = (),
        longopts: Tuple[str, ...] = ('--datadir',)) -> Optional[str]:
    for i, arg in enumerate(argv):  # pragma: no cover
        if arg in shortopts:
            try:
                value = argv[i + 1]
            except IndexError:
                import click
                raise click.UsageError(f'Missing value for {arg} option')
            return value
        if arg.startswith(longopts):  # pragma: no cover
            key, _, value = arg.partition('=')
            if not value:
                try:
                    value = argv[i + 1]
                except IndexError:
                    import click
                    raise click.UsageError(
                        f'Missing {key}! Did you mean --{key}=value?')
            return value
    return None


_datadir = (_extract_arg_from_argv(longopts=('--datadir',)) or
            os.environ.get('FAUST_DATADIR') or
            os.environ.get('F_DATADIR'))
if _datadir:  # pragma: no cover
    os.environ['FAUST_DATADIR'] = _datadir
_loop = (_extract_arg_from_argv(shortopts=('-L',), longopts=('--loop',)) or
         os.environ.get('FAUST_LOOP') or
         os.environ.get('F_LOOP'))
if _loop:  # pragma: no cover
    os.environ['FAUST_LOOP'] = _loop
    import mode.loop
    mode.loop.use(_loop)

# To ensure `import faust` executes quickly, this module imports
# attributes lazily. The next section provides static type checkers
# with information about the contents of this module.
if typing.TYPE_CHECKING:  # pragma: no cover
    from mode import Service, ServiceT                          # noqa: E402
    from .agents import Agent                                   # noqa: E402
    from .app import App                                        # noqa: E402
    from .auth import (                                         # noqa: E402
        GSSAPICredentials,
        SASLCredentials,
        SSLCredentials,
    )
    from .channels import Channel, ChannelT                     # noqa: E402
    from .events import Event, EventT                           # noqa: E402
    from .models import Model, ModelOptions, Record             # noqa: E402
    from .sensors import Monitor, Sensor                        # noqa: E402
    from .serializers import Codec, Schema                      # noqa: E402
    from .streams import Stream, StreamT, current_event         # noqa: E402
    from .tables.globaltable import GlobalTable                 # noqa: E402
    from .tables.table import Table                             # noqa: E402
    from .tables.sets import SetGlobalTable, SetTable           # noqa: E402
    from .topics import Topic, TopicT                           # noqa: E402
    from .types.settings import Settings                        # noqa: E402
    from .windows import (                                      # noqa: E402
        HoppingWindow,
        TumblingWindow,
        SlidingWindow,
        Window,
    )
    from .worker import Worker                                # noqa: E402
    from .utils import uuid

__all__ = [
    'Agent',
    'App',
    'AppCommand',
    'Command',
    'Channel',
    'ChannelT',
    'Event',
    'EventT',
    'Model',
    'ModelOptions',
    'Record',
    'Monitor',
    'Schema',
    'Sensor',
    'SetTable',
    'SetGlobalTable',
    'Codec',
    'Schema',
    'Service',
    'ServiceT',
    'Stream',
    'StreamT',
    'current_event',
    'Table',
    'GlobalTable',
    'Topic',
    'TopicT',
    'GSSAPICredentials',
    'SASLCredentials',
    'SSLCredentials',
    'Settings',
    'HoppingWindow',
    'TumblingWindow',
    'SlidingWindow',
    'Window',
    'Worker',
    'uuid',
]

# Lazy loading.
# - See werkzeug/__init__.py for the rationale behind this.
from types import ModuleType  # noqa

all_by_module: Mapping[str, Sequence[str]] = {
    'faust.agents': ['Agent'],
    'faust.app': ['App'],
    'faust.channels': ['Channel', 'ChannelT'],
    'faust.events': ['Event', 'EventT'],
    'faust.models': ['ModelOptions', 'Record'],
    'faust.sensors': ['Monitor', 'Sensor'],
    'faust.serializers': ['Codec', 'Schema'],
    'faust.streams': [
        'Stream',
        'StreamT',
        'current_event',
    ],
    'faust.tables.globaltable': ['GlobalTable'],
    'faust.tables.sets': ['SetTable', 'SetGlobalTable'],
    'faust.tables.table': ['Table'],
    'faust.topics': ['Topic', 'TopicT'],
    'faust.auth': [
        'GSSAPICredentials',
        'SASLCredentials',
        'SSLCredentials',
    ],
    'faust.types.settings': ['Settings'],
    'faust.windows': [
        'HoppingWindow',
        'TumblingWindow',
        'SlidingWindow',
        'Window',
    ],
    'faust.worker': ['Worker'],
    'faust.utils': ['uuid'],
    'mode.services': ['Service', 'ServiceT'],
}

object_origins = {}
for module, items in all_by_module.items():
    for item in items:
        object_origins[item] = module


class _module(ModuleType):
    """Customized Python module."""

    standard_package_vars = [
        '__file__',
        '__path__',
        '__doc__',
        '__all__',
        '__docformat__',
        '__name__',
        '__path__',
        'VERSION',
        'version_info_t',
        'version_info',
        '__package__',
        '__version__',
        '__author__',
        '__contact__',
        '__homepage__',
        '__docformat__',
    ]

    def __getattr__(self, name: str) -> Any:
        if name in object_origins:
            module = __import__(
                object_origins[name], None, None, [name])
            for extra_name in all_by_module[module.__name__]:
                setattr(self, extra_name, getattr(module, extra_name))
            return getattr(module, name)
        return ModuleType.__getattribute__(self, name)

    def __dir__(self) -> Sequence[str]:
        return sorted(list(new_module.__all__) + self.standard_package_vars)


# keep a reference to this module so that it's not garbage collected
old_module = sys.modules[__name__]

new_module = sys.modules[__name__] = _module(__name__)
new_module.__dict__.update({
    '__file__': __file__,
    '__path__': __path__,  # type: ignore
    '__doc__': __doc__,
    '__all__': tuple(object_origins),
    '__version__': __version__,
    '__author__': __author__,
    '__contact__': __contact__,
    '__homepage__': __homepage__,
    '__docformat__': __docformat__,
    '__package__': __package__,
    'version_info_t': version_info_t,
    'version_info': version_info,
    'VERSION': VERSION,
})



"""Faust exceptions."""
import typing

__all__ = [
    'FaustError',
    'FaustWarning',
    'FaustPredicate',
    'SecurityError',
    'NotReady',
    'Skip',
    'AlreadyConfiguredWarning',
    'ImproperlyConfigured',
    'ValidationError',
    'DecodeError',
    'KeyDecodeError',
    'ValueDecodeError',
    'SameNode',
    'ProducerSendError',
    'ConsumerNotStarted',
    'PartitionsMismatch',
    'ConsistencyError',
]

if typing.TYPE_CHECKING:
    from .types.models import FieldDescriptorT as _FieldDescriptorT
else:
    class _FieldDescriptorT: ...  # noqa


class FaustError(Exception):
    """Base-class for all Faust exceptions."""


class FaustWarning(UserWarning):
    """Base-class for all Faust warnings."""


class FaustPredicate(FaustError):
    """Base-class for semi-predicates such as :exc:`Skip`."""


class SecurityError(FaustError):
    """Base-class for security related (high priority) exceptions."""


class NotReady(FaustPredicate):
    """Service not started."""


class Skip(FaustPredicate):
    """Raised in stream processors to skip processing of an event."""


class AlreadyConfiguredWarning(FaustWarning):
    """Trying to configure app after configuration accessed."""


class ImproperlyConfigured(FaustError):
    """The library is not configured/installed correctly."""


class ValidationError(FaustError, ValueError):
    """Value passed for model field is not valid."""

    field: _FieldDescriptorT

    def __init__(self, reason: str, *,
                 field: _FieldDescriptorT) -> None:
        self.reason = reason
        self.field = field
        super().__init__(reason, field)

    def __str__(self) -> str:
        return f'{self.reason} {self.field!r}'

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self}>'


class DecodeError(FaustError):
    """Error while decoding/deserializing message key/value."""


class KeyDecodeError(DecodeError):
    """Error while decoding/deserializing message key."""


class ValueDecodeError(DecodeError):
    """Error while decoding/deserializing message value."""


class SameNode(FaustPredicate):
    """Exception raised by router when data is located on same node."""


class ProducerSendError(FaustError):
    """Error while sending attached messages prior to commit."""


class ConsumerNotStarted(NotReady):
    """Error trying to perform operation on consumer not started."""


class PartitionsMismatch(FaustError):
    """Number of partitions between related topics differ."""


class ConsistencyError(FaustError):
    """Persisted table offset is out of sync with changelog topic highwater."""



import faust
import pytest
from mode.utils.mocks import Mock, patch

app = faust.App('example-test-agent-call')


@app.agent()
async def foo(stream):
    async for value in stream:
        await bar.send(value)
        yield value


@app.agent()
async def bar(stream):
    async for value in stream:
        yield value + 'YOLO'


@app.agent()
async def baz(stream):
    async for value in stream:
        if value == 'hey':
            raise KeyError('foo')
        yield value


@pytest.fixture()
def test_app():
    app.finalize()
    app.conf.store = 'memory://'
    app.flow_control.resume()
    return app


def mock_coro(return_value=None, **kwargs):
    """Create mock coroutine function."""
    async def wrapped(*args, **kwargs):
        return return_value
    return Mock(wraps=wrapped, **kwargs)


@pytest.mark.asyncio()
async def test_foo(test_app):
    with patch(__name__ + '.bar') as mocked_bar:
        mocked_bar.send = mock_coro()
        async with foo.test_context() as agent:
            await agent.put('hey')
            mocked_bar.send.assert_called_with('hey')


@pytest.mark.asyncio()
async def test_bar(test_app):
    async with bar.test_context() as agent:
        event = await agent.put('hey')
        assert agent.results[event.message.offset] == 'heyYOLO'


@pytest.mark.asyncio()
async def test_agent_with_error(test_app):
    async with baz.test_context() as agent:
        event = await agent.put('foo')
        assert agent.results[event.message.offset] == 'foo'
        event = await agent.put('bar')
        assert agent.results[event.message.offset] == 'bar'
        with pytest.raises(KeyError):
            event = await agent.put('hey')
